#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>

#include "admutils.h"
#include "mcio.h"  

#define WVERSION   "310" 

// cdmin, gdmin added
// fmname, lmname added  (all pairs in region)

#define MAXFL  50   
#define MAXSTR  512

char *trashdir = "/var/tmp" ;
int verbose = NO ;
int qtmode = NO ;
Indiv **indivmarkers;
SNP **snpmarkers ;
int numsnps, numindivs ; 

double plo = .001 ;
double phi = .999 ;

double reichthresh = -1.0 ;
// this is a threshold (p value) at 1 cM 
// we alter it linearly  so reichthresh of .002 at 1cM means thresh of .02 at .1 cM  
// We need to threshold in the code with a zscore, and our LRT is 2 sided 
// so we need to divide by 2 before calculating zthresh

char  *genotypename = NULL ;
char  *snpname = NULL ;
char  *snpoutfilename = NULL ;
char  *indivname = NULL ;
char *badsnpname = NULL ;

char *eglistname  = NULL ;
char *outputname = NULL ;
FILE *ofile ;
double cdmin = -1.0e6; 
double gdmin = -1.0e6; 

char *fmname = NULL ;
char *lmname = NULL ;
/** look at all markers between fmname, lmname */

char  unknowngender = 'U' ;
int   xchrom = -1 ;
int maxlag = 4 ;

void readcommands(int argc, char **argv) ;

int mkeglist(char **eglist);
void seteglist(Indiv **indm, int nindiv, char *eglistname);

int snpindx(char *snpid) ;

void cntit(double *xc, SNP *c1, SNP *c2) ;
void cntitx(double *xc, double *xd, SNP *c1, SNP *c2) ;

double lddip(double *xc) ;
double lddipx(double *xc, double *xd) ;

int main(int argc, char **argv)
{

  int **snppos ;
  int *snpindx ;
  char **snpnamelist, **indnamelist ;
  char **eglist ;
  int  lsnplist, lindlist, numeg ;
  int i,j; 
  SNP *cupt, *cupt1, *cupt2, *cupt3 ;
  Indiv *indx ;
  double gpos1,gpos2,cpos1,cpos2,gd, cd, gd100 ;
  double rthresh, zt ;
  int mpflag ;

  int ch1, ch2 ;
  int fmnum , lmnum ;
  int num, n1, n2 ;

  int nindiv = 0, e, f, lag=1  ;
  double xc[9], xd[4], xc2[9] ;
  double ychi, zscore, zthresh = 20.0 ;
  int nignore, numrisks = 1 ;

  readcommands(argc, argv) ;
  ofile = stdout; 
  if (outputname != NULL) openit(outputname, &ofile, "w") ;

  if (reichthresh>0.0) { 
    fprintf(ofile, "##reichhresh: %9.3f\n", reichthresh) ;
  }
  if (gdmin>0.0) { 
    fprintf(ofile, "##gdmin: %9.3f\n", gdmin) ;
  }
  if (cdmin>0.0) { 
    fprintf(ofile, "##cdmin: %9.0f\n", cdmin) ;
  }



  numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, badsnpname, &nignore, numrisks) ;

  numindivs = getindivs(indivname, &indivmarkers) ;
  getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;

  printsnps(NULL, snpmarkers, numsnps, 
   indivmarkers, NO, NO) ;

  ZALLOC(eglist, numindivs, char *) ; 
  numeg = mkeglist(eglist) ;
  for (i=0; i<numeg; i++) {  
   printf("%3d %s\n",i, eglist[i]) ;
  }
  seteglist(indivmarkers, numindivs, eglistname);
  nindiv=0 ;
  for (i=0; i<numindivs; i++) {
    indx = indivmarkers[i] ;
    if(indx -> affstatus == YES) ++nindiv  ;
  }
 printf("number of individuals in analysis %d\n", nindiv) ;

 if (fmname != NULL) maxlag = 0 ;

 for (lag=1; lag <=maxlag; ++lag)  {
  for (i=0; (i+lag)<numsnps ; i++) { 
   cupt1 = snpmarkers[i] ;
   cupt2 = snpmarkers[i+lag] ;
   if (cupt1 -> ignore) continue ;
   if (cupt2 -> ignore) continue ;
   mpflag = NO ;
   ch1 = cupt1 -> chrom ;
   ch2 = cupt2 -> chrom ;
   if (ch1==0) fatalx("bad marker (chrom 0) %s\n", cupt1 -> ID) ;
   if (ch1 != ch2) continue ;
   if ((xchrom>0) && (xchrom != ch1)) continue ;
   if (ch2 < 23) {
    cntit(xc, cupt1, cupt2) ;
    ychi = lddip(xc) ;
   }
   else {
    cntitx(xc, xd, cupt1, cupt2) ;
    ychi = lddipx(xc, xd) ;
   }
   if (ychi< -0.001) {
    printf(" badlddip %9.3f\n",  ychi) ;
    printmat(xc, 3, 3) ;
    fatalx("badlddip\n") ;
   } 
   ychi = MAX(ychi, 1.0e-10) ;
   zscore = sqrt(ychi) ;
   gpos1 = cupt1 -> genpos ;
   cpos1 = cupt1 -> physpos ;
   gpos2 = cupt2 -> genpos ;
   cpos2 = cupt2 -> physpos ;
   gd = gpos2 -gpos1 ;
   if (gd < gdmin) mpflag = YES  ;
   cd = cpos2 -cpos1 ;
   if (cd < cdmin) mpflag = YES ;

   cpos1 /= 1.0e6 ;
   cpos2 /= 1.0e6 ;
   cd /= 1.0e6 ;

   if (reichthresh>0.0) {  
    gd100 = gd * 100.0 ; //cM
    gd100 += 1.0e-5 ;
    rthresh = 0.5*reichthresh/gd100 ;
    if (rthresh<0.5) {
     zt = zprob(rthresh) ;
    }
    else {  
     zt = -1.0 ;
    }
    printf("zzq: %4d %4d %9.3f %9.3f %9.3f %9.3f\n", i, lag, gd100, rthresh, zt, zscore) ;
    if (zscore >  zt) mpflag = YES ; 
   } // end reichthresh

   if (mpflag == NO) continue ;

   fprintf(ofile,"%4d %4d %4d",i, i+lag, ch1) ;
   fprintf(ofile," %18s %18s", cupt1->ID , cupt2->ID) ; 
   fprintf(ofile," %9.3f", zscore) ;
   fprintf(ofile, " %2d %9.3f %9.3f", lag, gpos1, cpos1) ;
   fprintf(ofile, " %9.3f %9.3f", gd, cd) ;
   fprintf(ofile, "\n") ; 
   
   if (zscore>zthresh) {
    printf("%4d %4d %4d",i, i+lag, ch1) ;
    printf(" %15s %15s", cupt1->ID , cupt2->ID) ; 
    printf(" %9.3f\n", zscore) ;
    printmat(xc, 3, 3) ;
   }
  }
 }
 if (fmname != NULL) {  
   fmnum = snpindex(snpmarkers, numsnps, fmname) ;
   if (fmnum<0) fatalx("markername %s not found\n", fmname) ;
   lmnum = snpindex(snpmarkers, numsnps, lmname) ;
   if (lmnum<0) fatalx("markername %s not found\n", lmname) ;

   for (i=fmnum; i<= lmnum; i++)  {  
    for (j=i+1; j<= lmnum; j++)  {  

    cupt1 = snpmarkers[i] ;
    cupt2 = snpmarkers[j] ;
    if (cupt1 -> ignore) continue ;
    if (cupt2 -> ignore) continue ;
    mpflag = NO ;
    ch1 = cupt1 -> chrom ;
    ch2 = cupt2 -> chrom ;
    if (ch1==0) fatalx("bad marker (chrom 0) %s\n", cupt1 -> ID) ;
    if (ch2==0) fatalx("bad marker (chrom 0) %s\n", cupt2 -> ID) ;
    if (ch2 < 23) {
     cntit(xc, cupt1, cupt2) ;
     ychi = lddip(xc) ;
     num = nnint(asum(xc,9)) ;
    }
    else {
     cntitx(xc, xd, cupt1, cupt2) ;
     ychi = lddipx(xc, xd) ;
     num = nnint(asum(xc,9)) ;
     num += nnint(asum(xd,4)) ;
    }
    if (ychi< -0.001) {
     printf(" badlddip %9.3f\n",  ychi) ;
     printmat(xc, 3, 3) ;
     fatalx("badlddip\n") ;
    } 
    ychi = MAX(ychi, 1.0e-10) ;
    zscore = sqrt(ychi) ;
    gpos1 = cupt1 -> genpos ;
    cpos1 = cupt1 -> physpos ;
    gpos2 = cupt2 -> genpos ;
    cpos2 = cupt2 -> physpos ;
    gd = gpos2 -gpos1 ;
    cd = cpos2 -cpos1 ;

    cpos1 /= 1.0e6 ;
    cpos2 /= 1.0e6 ;
    cd /= 1.0e6 ;

    lag = j-i ;

    fprintf(ofile,"%4d %4d %4d",i, j, ch1) ;
    fprintf(ofile," %18s %18s", cupt1->ID , cupt2->ID) ; 
    fprintf(ofile," %9.3f", zscore) ;
    fprintf(ofile, " %2d %9.3f %9.3f", lag, gpos1, cpos1) ;
    fprintf(ofile, " %9.3f %9.3f", gd, cd) ;
    fprintf(ofile, " %4d", num) ;
    fprintf(ofile, "\n") ; 
   }
  }
 }
  fclose(ofile) ;
  printf("##end of run\n") ;
  return 0 ;
}



void cntit(double *xc, SNP *c1, SNP *c2) 
{
 int n, i, e, f ;
 Indiv *indx ;
 n = MIN(c1->ngtypes, c2->ngtypes) ;
 vzero(xc, 9) ;
 for (i=0; i<n; i++) {  
  indx = indivmarkers[i] ;
  if (indx->affstatus != YES) continue ;
  e = c1->gtypes[i] ;
  f = c2->gtypes[i] ;
  if (e<0) continue ;
  if (f<0) continue ;
  ++xc[3*e+f] ;
 }
}

void cntitx(double *xc, double *xd, SNP *c1, SNP *c2) 
{
 int n, i, e, f ;
 Indiv *indx ;
 n = MIN(c1->ngtypes, c2->ngtypes) ;
 vzero(xc, 9) ;
 vzero(xd, 9) ;
 for (i=0; i<n; i++) {  
  indx = indivmarkers[i] ;
  if (indx->affstatus != YES) continue ;
  if (indx->gender == 'F') {
   e = c1->gtypes[i] ;
   f = c2->gtypes[i] ;
   if (e<0) continue ;
   if (f<0) continue ;
   ++xc[3*e+f] ;
  }
  if (indx->gender == 'M') {
   e = c1->gtypes[i] ;
   f = c2->gtypes[i] ;
   if (e<0) continue ;
   if (f<0) continue ;
   e = e/2 ;
   f = f/2 ;
   ++xd[2*e+f] ;
  }
 }
}

int mkeglist(char **eglist)
{
  Indiv *indx ;
  int i, k, numeg=0 ;
  for (i=0; i<numindivs; i++) { 
   indx = indivmarkers[i] ;
    k = indxindex(eglist, numeg,  indx->egroup) ;
    if (k<0)   { 
     eglist[numeg] = strdup(indx->egroup) ;
     ++numeg ;
    }
  }
  return numeg ;
}

void checkgeno() 
{
  int i, k, x ;
  int n, nbad, ngood ;
  SNP *cupt ;

  nbad = ngood = 0 ;
  for (i=0; i<numsnps; i++) {  
   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   n = cupt -> ngtypes ;
   if (n<=0) continue ;
   for (k=0; k<n; k++) {
    x = cupt -> gtypes[k] ;
    if (x>2) {  
     printf("bad genotype: %s %4d %4d\n", cupt->ID, k,x) ;
     cupt -> gtypes[k]  = -1 ;
     ++nbad ;
    }
    else  {
     ++ngood ;
    }
   }
  }
  printf("checkgeno. good: %d  bad:  %d\n",ngood, nbad) ;
/**
  if (nbad>0) fatalx("checkgeno fails\n") ;
*/
}

void
countgtypes(SNP *cupt, int *xcnt)  
{
   int k, numstates, ignore, refcnt, varcnt, a ;
   int maxcnt ;
   Indiv *indx ;

   ivzero(xcnt, 4) ;
   if (cupt -> isfake) return ;
   if (cupt -> ignore) return ;
   if (cupt -> gtypes == NULL) return ;
   if (cupt -> ngtypes == 0) return ;
   for (k=0; k< numindivs; ++k) {  
    indx = indivmarkers[k];
    gettln(cupt, indx, NULL, NULL, &numstates, &ignore) ;
    if (ignore == YES) continue ;
    a = 0 ;
    if (indx -> affstatus) a = 2 ;  
    refcnt = cupt -> gtypes[k] ;
    if (refcnt<0) continue ;
    maxcnt = 2  ;
    if (numstates==2) {  
     maxcnt = 1 ;
     refcnt /= 2 ;
    }
    varcnt = maxcnt-refcnt ;
    xcnt[a] += refcnt ;
    xcnt[a+1] += varcnt ;
   }
}

void readcommands(int argc, char **argv) 

{
  int i,haploid=0;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

/**
DIR2:  /fg/nfiles/admixdata/ms2
SSSS:  DIR2/outfiles 
genotypename: DIR2/autos_ccshad_fakes
eglistname:    DIR2/eurlist  
output:        eurout
*/
   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "eglistname:", &eglistname) ;
   getstring(ph, "output:", &outputname) ;
   getint(ph, "chrom:", &xchrom) ;
   getint(ph, "maxlag:", &maxlag) ;
   getdbl(ph, "reichthresh:", &reichthresh) ;
   getdbl(ph, "cdmin:", &cdmin) ;
   getdbl(ph, "gdmin:", &gdmin) ;
   getstring(ph, "fmname:", &fmname) ;
   getstring(ph, "lmname:", &lmname) ;
   getstring(ph, "badsnpname:", &badsnpname) ;

}

void setfliplist(char *flistname, char **flist, int *numfl)
{
  FILE *fgfile ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, num=0 ;

  *numfl = 0; 
  if (flistname == NULL) return ;
  openit(flistname, &fgfile, "r") ;
  while (fgets(line, MAXSTR, fgfile) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') continue ;
   if (num>=MAXFL) 
  fatalx("too many flips\n") ;
   flist[num] = strdup(sx) ;
   ++num ;
   freeup(spt, nsplit) ;
  }
  fclose(fgfile) ;
  *numfl = num ;
}

void seteglist(Indiv **indm, int nindiv, char *eglistname) 
{
  FILE *egfile ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  Indiv *indx ;
  int nsplit, i ;

  if (eglistname == NULL) return ;
  openit(eglistname, &egfile, "r") ;
  while (fgets(line, MAXSTR, egfile) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   setstatus(indm, nindiv, sx) ;
   freeup(spt, nsplit) ;
  }
  fclose(egfile) ;
}


int snpindx(char *snpid) 
/* hash table would be good here */
{
     int k ; 
     for (k=0; k< numsnps; k++) {  
      if (strcmp(snpmarkers[k] -> ID, snpid) == 0) return k ;
     }
     return -1 ;
}

void flipallele(SNP *cupt) 
// flip reference, variant counts
{
   int i, n, g ;
   n = cupt -> ngtypes ;
   for (i=0; i<n; i++) {  
    g = cupt -> gtypes[i] ;
    if (g<0) continue ;
    cupt ->gtypes[i] = 2-g ;
   }
}
