#include "mclm.h" 
#include <math.h>

#define NUMX  13 

// if NUMX is odd lam 1, mu 1 are in model 

extern int markernum ; 
extern int lmnumx    ;
extern int oldlmmode ;
extern double lmmax  ;
extern double lmthresh ;
extern  int lmdetails ;
extern double lrisk ;

double lol, hil, lom, him ;
static int numlm, xchrom  ;  
static double *lamx, *mux ;
static double **xpsi ;

static void xmul(double *z, double *a, double *b) ;

void setlm(SNP **snpm, int numsnps, int chrom)  
{
   int i, j, k, c, d ;
   double a, b, inc ;
   double lam, mu, xlam, xmu, *pp ;
   SNP *cupt ;

   if (lmnumx<0) lmnumx = NUMX ;
   printf("setlm:  lmnumx:  %d  lmmax:  %9.3f\n", lmnumx, lmmax) ;
   printf("(mclm) riskarr: %9.3f\n", lrisk) ; 
   b = log(lmmax) ;
   a = -b ;
   inc = (b-a) / (double) (lmnumx-1) ;
   numlm = lmnumx*lmnumx ;
   ZALLOC(lamx, numlm, double) ; 
   ZALLOC(mux, numlm, double) ; 
   ZALLOC(xpsi, numlm, double *) ; 
   lol = lom  = a ;
   k = 0 ;
   for (i=0; i<lmnumx ; i++)  {
    for (j=0; j<lmnumx ; j++)  {
     lam = lol + inc * (double) i  ;
     mu = lom + inc * (double) j  ;
     xlam = lamx[k] = exp(lam) ;
     xmu = mux[k] = exp(mu) ;
     ZALLOC(xpsi[k], 9, double) ;
     pp = xpsi[k] ; 
     ++k ;
     for (c=0; c<3 ; c++) {
      for (d=0; d<3 ; d++) {
       pp[3*c+d] = pow(xlam, (double) c) * pow(xmu, (double) d) ;
      }
     }
    }
   }
   xchrom = chrom ;
   for (i=0; i<numsnps; i++) {
    cupt = snpm[i] ;
    if (ignoresnp(cupt)) continue ;
    if ((xchrom>0) && (cupt->chrom != xchrom)) continue ;
    if (cupt -> modelscores != NULL) { 
     free(cupt->modelscores) ;
    }
    if (cupt -> totmodelscores != NULL) { 
     free(cupt->totmodelscores) ;
    }
    ZALLOC(cupt -> modelscores, numlm, double) ;
    ZALLOC(cupt -> totmodelscores, numlm, double) ;
   }
   if (oldlmmode == NO) { 
    numlm = lmnumx ;
    for (j=0; j<lmnumx ; j++)  {
     mu = lom + inc * (double) j  ;
     mux[j] = exp(mu) ;
    }
   }
}
void clearmodelscores(SNP **snpm, int numsnps, int mode) 
{
  int i ;
  SNP *cupt ;

   for (i=0; i<numsnps; i++) {
    cupt = snpm[i] ;
    if (ignoresnp(cupt)) continue ;
    if ((xchrom>0) && (cupt->chrom != xchrom)) continue ;
    vzero(cupt -> modelscores, numlm) ;
    if (mode)
     vzero(cupt -> totmodelscores, numlm) ;
   }

}

void mkbprob(double *bprob, double popAfreq, double popBfreq, int numstates) 
// bprob[a+numstates + b] is P(geno=b | ancestry = a) 
{
  int a ;
  double XA[2], XC[2] ;
  XA[0] = 1.0 - popAfreq ;
  XA[1] = popAfreq ;
  XC[0] = 1.0 - popBfreq ;
  XC[1] = popBfreq ;
  vzero(bprob, 9) ;
  if (numstates == 2) { 
   copyarr(XA, bprob, 2) ;
   copyarr(XC, bprob+3, 2) ;
   return ;
  }
  xmul(bprob, XA, XA) ; 
  xmul(bprob+3, XA, XC) ; 
  xmul(bprob+6, XC, XC) ; 
}
static void xmul(double *z, double *a, double *b) 
// multiply 2 degree 1 polys !!
{
  z[0] = a[0]*b[0] ;
  z[1] = a[0]*b[1] +a[1]*b[0] ;
  z[2] = a[1]*b[1] ;
  
}

void calclm(SNP *cupt , Indiv *indx, double *gg, double *ypsi) 
// lmscoring  case only of course
{

     int a, b, bb, idnum, k ;
     double theta, lambda ;
     double bprob[9] ;
     double tt[3], top, bot, pp, y ;
     int numstates, ignore ;
     static int ncall = 0 ;
     int mnum ;
     double xlam, xmu,  mm[3], xxpsi[3] ;

     if (ignoresnp(cupt)) return ;
     gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ;
     if (indx -> affstatus == NO) return ;
     if (ignore) return ;
     if ((xchrom>0) && (cupt->chrom != xchrom)) return ;
     idnum = indx -> idnum ;
     bb = cupt -> gtypes[idnum] ;
     if (bb<0) return ;
      ++ncall ;
     if (numstates==2) bb /= 2 ;
     mkbprob(bprob, cupt -> adm_PopA_freq, cupt -> adm_PopB_freq, numstates) ;
     sett1(tt, theta, numstates) ;
     if (oldlmmode == NO) {  
	      for (k=0; k<numlm; ++k) {   
	       xmu = mux[k]  ;

	       mm[0] = 1.0 ;
	       mm[1] = xmu ;
	       mm[2] = xmu*xmu ;

	       for (a=0; a<numstates; a++) {
		bot = vdot(bprob+3*a, mm, numstates) ;
		top = ypsi[a]*vdot(bprob+3*0, mm, numstates) ;
		xlam = top/bot  ;   
		xxpsi[a] = xlam*mm[bb] ;
	       }
	       top = vdot(gg, xxpsi, numstates) ;
	       bot = vdot(gg, ypsi, numstates) ;
	       cupt -> modelscores[k] += log(top/bot) ;
	      }
	      return ;
     }
	     for (k=0; k<numlm; k++) {  
	      top = bot = 0.0 ;
	      for (a=0; a<numstates; a++) {
	       for (b=0; b<numstates; b++) {
		pp = xpsi[k][3*a+b] ;
		bot += tt[a]*bprob[3*a+b]*pp ;
		if (b==bb) top += gg[a]*pp ;
	       }
	      }
	      cupt -> modelscores[k] += log(top/bot) ;
	     }
}
void addmodelscores(SNP **snpm ,  int numsnps) 
{

   int i ;
   SNP *cupt ;
   double *xx ;
   static int ncall = 0 ;
   double xhi ;  

   ++ncall ;
   ZALLOC(xx, numlm, double) ;
   for (i=0; i<numsnps; i++) {
    cupt = snpm[i] ;
    if (ignoresnp(cupt)) continue ;
    if ((xchrom>0) && (cupt->chrom != xchrom)) continue ;
    xhi = 40.0*log(10.0) ;
    vclip(xx, cupt -> modelscores, -xhi, xhi, numlm) ;
    vexp(xx, xx, numlm) ;
    vvp(cupt -> totmodelscores, cupt -> totmodelscores, 
      xx, numlm) ;

    if ((ncall<=-1) && (i<=10)) {
     printf("zz %d %s\n", ncall, cupt ->ID) ;
     printmat(cupt -> modelscores, lmnumx, lmnumx) ;
     printf("zz2\n") ;
     printmat(cupt -> totmodelscores, lmnumx, lmnumx) ;
    }

   }
   free(xx) ;
}

void printmodelscores(double *xscore, SNP **snpm,  int numsnps, int numiter) 
{
   int i, k ;
   SNP *cupt ;
   double *xx ;
   double y , yfac, ylfac, y1, y2 ;
   double lbest, mbest, sbest, lmscbest ;
   double yafac, ybfac, lmscore ;
   int na, nb, chrom, pos ;
   char bestID[100] ;  
   double lmbayes, lmweight ;
   double bprob[9], mm[3], xlam, xmu;
   double top, bot ;
   int a, numstates = 3 ; 

   printf("\n\n") ;
   ZALLOC(xx, numlm, double) ;
   y = (double) numiter ;
   lmscbest = -1000.0 ;
   lmbayes = 0.0 ;
   lmweight = 0.0 ; // for global score */

   for (i=0; i<numsnps; i++) {
    cupt = snpm[i] ;
    if (ignoresnp(cupt)) continue ;
    chrom = cupt -> chrom ;
    pos = cupt -> physpos ;
    if ((xchrom>0) && (chrom != xchrom)) continue ;
    vst(xx, cupt -> totmodelscores, 1.0/y, numlm) ;
    vsp(xx, xx, 1.0e-8, numlm) ;
    yfac = asum(xx, numlm) / (double) numlm ;
    lmscore = ylfac = log10(yfac) ;
    lmscbest = MAX(lmscbest, ylfac) ;
    if ((ylfac<lmthresh) && (! cupt -> isrfake)) continue ; 
    mkbprob(bprob, cupt -> adm_PopA_freq, cupt -> adm_PopB_freq, numstates) ;
    y1 = pow10(xscore[i]) ;
    y2 = yfac ; 
    lmweight += y1 ;
    lmbayes  += y1*y2 ;
    if (oldlmmode == NO) {  
      vlog10(xx, xx, numlm) ;
      for (k=0; k<numlm; ++k) {
       if (lmdetails) 
        xmu = mux[k] ;
        printf("lmdetails %3d %9.3f %9.3f ", k,  mux[k], xx[k]) ;
        mm[0] = 1.0 ;
        mm[1] = xmu ;
        mm[2] = xmu*xmu ;
        a = 1 ;  
        top = lrisk * vdot(bprob+3*0, mm, numstates) ;
	bot = vdot(bprob+3*a, mm, numstates) ;
	xlam = top/bot ;  
        printf("  %9.3f\n", xlam) ;
      }
      printf("lmscore: %15s %9.3f", cupt -> ID, lmscore) ;
      printf("%3d %15d %9.3f\n", chrom, pos, xscore[i]) ;
      continue ;
    }
    sbest = -1000 ; 
    strcpy(bestID, "NULL") ;
    yafac = ybfac = 0.0 ;
    na = nb = 0 ;
    vlog10(xx, xx, numlm) ;
    for (k=0; k<numlm ; ++k) {  
     ylfac = xx[k] ;
     if (fabs(mux[k]-1.0) < 0.01) {
       ybfac += pow10(ylfac) ; ++nb ;
     }
     if (fabs(lamx[k]-1.0) < 0.01) {
       yafac += pow10(ylfac) ; ++na ;
     }
     if (lmdetails) 
      printf("lmdetails %3d %9.3f %9.3f %9.3f\n", k, lamx[k], mux[k], ylfac) ;
     if (ylfac>sbest) { 
      sbest = ylfac ;
      lbest = lamx[k] ;
      mbest = mux[k] ;
      strncpy(bestID, cupt -> ID, 99) ;
     }
    }
    printf("lmsctop: %9.3f %9.3f %9.3f %s\n", lbest, mbest, sbest, bestID) ;
    if ((na>0) && (nb>0)) {
// first number is averaging over risk for allele only 
// second is ancestry only 
      yafac /= (double) na ;
      ybfac /= (double) nb ;
      printf("lmabscores: %9.3f %9.3f\n", log10(yafac), log10(ybfac) ) ;
      lmscore -= log10(ybfac) ;
      printf("lmscore: %15s %9.3f", cupt -> ID, lmscore) ;
      printf("%3d %15d %9.3f\n", chrom, pos, xscore[i]) ;
    }
   }
   free(xx) ;
   printf("lmscbest: %9.3f\n", lmscbest) ;
   lmbayes /= lmweight ;
   printf("lmbayes:  %9.3f\n", log10(lmbayes) ) ;
   printf("\n\n") ;
}
