#include "mcsubs.h" 
#include "qsubs.h" 
#include "gammas.h"

double risk ;

extern int ncontrols ;
extern int leave1mode ;

int numindivs ;

Indiv *indiv_array ;

void unlinktemp()  ;

void loadgams(FILE *gamonefile,int markernum, double *gammas, int *numindivs, SNP *marker);

void floadgams(int ff,int markernum, double *gammas, 
  int *numindivs, SNP *marker);

int opengamfile(char *gamname, int *nummark, int *numindivs) ;

void unlinkgf() ;
void settt(Indiv *array, int numloaded) ;

double neil_score
 (Indiv *indiv_array, double **gams,
  int numindivs, double risk, int chrom);

double nick_score
 (Indiv *indiv_array, double **gams,
  int numindivs, double risk, int chrom);

double
checkgams(Indiv *indiv_array, 
  double **alphas, double **betas, double **gammas, 
  int numindivs)  ;

double mapcheckx(SNP *cupt)  ;
double freqcheckx(SNP *cupt, double *rp, double *rpp, 
 int cmode, int vmode)  ;
void doleave1(double *topscore, double *botscore, 
  SNP  *cupt, int lrmode, int numsnps, int winwidth)  ;
void checkindiv (SNP **snpm, Indiv **indm, int numsnps, int numindivs)  ;

void grabvx(SNP *cupt, double *vv, int *np)  ;

double
reest(int *obs, double **gg, 
 int numindiv, double *xp, double *xpp, int *ismalex)  ;

void docalcXX(int x, double *X, double **XX,
 double p, double pp, int numstates) ;

void
copywingams(double ***wingams, double **gammas, int n, int col)  ;

void pubscores (double *basescore, double *topscore, double *botscore,
 SNP **snpm, int numsnps) ;

static int 
isitpeak(SNP **snpm,  int k, double winlen, double thresh) ;  

int  nextfake (int start, int end, int inc) ;

double HMMwindow (
 
 SNP **snpm, 
 int indiv,
 double theta, 
 double lambda, 
 int first, 
 int last, 
 int leave1out,  /* -1 if none */
 double **gammas,
 double *ainit, 
 double *binit, 
 int numstates );
/* returns log10 score */


#define NCHROM 23 
#define NCHROMP (NCHROM+1)

void leave1(SNP **snpmarkers, int numsnps, int numind, 
 double lrisk, char *scratchdir)
{
    int numloaded, i, chrom ;

    double **alphas, **betas, **gammas ;
    SNP *cupt ;
    double *basescore, *neilsc ;
    double fac, totfac = 0.0 ;
    double *topscore, *botscore ;
    int numsc = 0 ;
    double bestsc = -1000.0 ;
    double bestscore[NCHROMP] ;
    double bestphys[NCHROMP] ;
    double bestpos[NCHROMP] ;
    double bestmark[NCHROMP] ;
    FILE *scorefile ;
    int numstatesg = 3 ;
    double p0, pp0, tmp, tmp2, p1, pp1 ;
    int winwidth = 20 ; /* leave1 */
    int f0, f1, f2 ;

    printf("###Entering leave1 risk: %9.3f\n\n", lrisk) ;

    numindivs = numind ;
    risk = lrisk ;

// clean up at some point.  
    ZALLOC(indiv_array, numindivs, Indiv) ;
    for (i=0; i<numindivs; i++) {  
     indiv_array[i] = *indivmarkers[i] ;
    }

    ZALLOC(neilsc, numsnps, double) ;
    ZALLOC(basescore, numsnps, double) ;
    ZALLOC(topscore, numsnps, double) ;
    ZALLOC(botscore, numsnps, double) ;


    vclear(bestscore, -1000, NCHROMP) ;
    scorefile = stdout ;

    dumpgams(snpmarkers, numsnps, numindivs, scratchdir) ;

    gammas = initarray_2Ddouble(numstatesg,numindivs,0.0);
    alphas = initarray_2Ddouble(numstatesg,numindivs,0.0);
    betas = initarray_2Ddouble(numstatesg,numindivs,0.0);

    numloaded = numindivs ;
    getfd(&f0, &f1, &f2) ;

  for (i=0; i<numsnps; i++) {

    cupt = snpmarkers[i] ;  

    floadgams(f0,i,gammas[0],&numloaded,cupt);
    floadgams(f1,i,gammas[1],&numloaded,cupt);
    floadgams(f2,i,gammas[2],&numloaded,cupt);

    chrom = cupt -> chrom ;

    tmp = basescore[i] =
     nick_score(indiv_array, gammas, numindivs, risk, chrom);

    neilsc[i] = 
     neil_score(indiv_array, gammas, numindivs, risk, chrom);

    if (cupt->ignore) continue ;
    tmp = MIN(tmp, 20.0) ;  /* stop overflow */  
    fac = pow10(tmp) ;

    if (cupt -> isfake) {
     totfac += fac ;
     ++numsc ;
    }

    if (verbose)  { 
     printf("%20s %5d  %9.3f \n",cupt -> ID, i, tmp) ;
    }

    bestsc = MAX(bestsc, basescore[i]) ;

    if (basescore[i]>bestscore[chrom])  {  
     bestscore[chrom] = basescore[i] ;
     bestpos[chrom] = cupt -> genpos ;
     bestphys[chrom] = cupt -> physpos ;
     bestmark[chrom] = i ;
    }
    if (verbose) {
     fprintf(scorefile,"%20s %5d %9.3f",cupt -> ID, i, basescore[i]) ;
     fprintf(scorefile, " %9.3f", neilsc[i]) ;
     fprintf(scorefile," %2d %12.0f %12.3f", 
      chrom, cupt->physpos, cupt->genpos) ;
     fprintf(scorefile,"\n") ;
    }

  }

/* mapcheck */
  printf("\n###MAPCHECK RESULTS FOLLOW:\n");
  printf("##%30s %10s %10s %12s\n", "SNP_ID","SNP_INDEX", "CHROM", "ANCESTRY_DIFF"); 
  for (i=0; i<numsnps; i++) {
    cupt = snpmarkers[i] ;  

    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;

    tmp = mapcheckx(cupt) ; 
    
    printf("mapcheck %25s %8d %10d %11.3f\n", cupt -> ID, i,  cupt -> chrom, tmp) ;
   }

/* freqcheck */
  printf("\n###FREQCHECK RESULTS FOLLOW:\n");
  printf("##For information about this check, look at the documentation included\n");
  printf("##%28s %11s %11s %10s %9s %9s %12s %12s\n","SNP_ID","CHR_NUM","S(ALL)","S(CONTROLS)","F(A)","F(E)","G(A)","G(E)");

 
  tmp2 = -1.0 ;
  for (i=0; i<numsnps; i++) {
    cupt = snpmarkers[i] ;  

    chrom = cupt -> chrom ;
    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;

    p0 = cupt -> adm_PopA_freq ;
    pp0 = cupt -> adm_PopB_freq ;
    tmp =  2.0*freqcheckx(cupt, &p1, &pp1,2, NO) ;  /* everybody */
    if (ncontrols>0)  
     tmp2 = 2.0*freqcheckx(cupt, &p1, &pp1,0, NO) ; /* controls  */
/**
  This test checks to see if we can get an improved
  fit to the observations from different allele 
  frequencies.  The test is perhaps a bit dubious since 
  we are fitting to frequencies obtained from the MCMC.
 
  Perhaps we should try naive frequencies from the 
   modern populations too.
*/
    printf("freqcheck %20s %8d %11.3f %11.3f", cupt -> ID, chrom, tmp, tmp2) ;
    printf(" %11.3f %11.3f   %11.3f %11.3f\n", p0, pp0, p1, pp1) ;
   }

   copyarr(basescore, topscore, numsnps) ;
   copyarr(basescore, botscore, numsnps) ;

   for (i=0; i<numsnps; i++) {
    
    if (leave1mode == NO) break ;

    cupt = snpmarkers[i] ;  

    chrom = cupt -> chrom ;

    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;

    doleave1(topscore, botscore, cupt,0, numsnps, winwidth) ; 
     /* marker on left */
    doleave1(topscore, botscore, cupt,1, numsnps, winwidth) ; 
     /* marker on right */
   }

   pubscores(basescore, topscore, botscore, snpmarkers, numsnps) ;
  

  free2D(&alphas, numstatesg) ;
  free2D(&betas, numstatesg) ;
  free2D(&gammas, numstatesg) ;
  free(basescore) ;
  free(topscore) ;
  free(botscore) ;
  free(neilsc) ;
  free(indiv_array) ;

  totfac /= (double) numsc ;
  printf("global score (leave1): %9.3f\n", log10(totfac)) ;

    unlinktemp() ;
    printf(" exiting leave1\n") ;
}

double neil_score
 (Indiv *indiv_array, double **gams, int numindivs, 
 double xrisk, int chrom)
/* zstat */
{

  SNP *cupt ;
  Indiv *indx ;

  double ee[3], gg[3] ;
  int numstates, i, fc ;
  int ignore ;
  double theta, lambda, xcase, xcontrol ;

  double ynumx, tmp ;
  double z1case, z2case, z1control, z2control, w1, w2 ;
  


  ee[0] = 0; ee[1] = 1; ee[2] = 2 ;

  xcase = xcontrol = 0 ;

  z1case = z2case = 0 ;
  z1control = z2control = 0 ;

  for (i=0; i<numindivs; i++) {
    gg[0] = gams[0][i] ;
    gg[1] = gams[1][i] ;
    gg[2] = gams[2][i] ;
    fc = fchrom[chrom] ;
    cupt = snpmarkers[fc] ;
    indx = indiv_array + i ;
    gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ;

   ynumx = (double) numstates -1 ;

   tmp = vdot(gg, ee, numstates) / ynumx  ;
   w1 = tmp -= theta ;

   if (indx -> affstatus)  {  
    z1case += w1 ;
    z2case += w1 * w1 ;
    ++xcase ;
   }
   else {
    z1control += w1 ;
    z2control += w1 * w1 ;
    ++xcontrol ;
   }

  } 

  /*AT: 04/08/05: If there are no cases or controls we should not calculate this*/
  if(xcase == 0 || xcontrol == 0)
    return 0.0;

   z1case /= xcase ;
   z2case /= xcase ;
   z2case -= (z1case*z1case) ;
   
   z1control /= xcontrol ;
   z2control /= xcontrol ;
   z2control -= (z1control*z1control) ;
   
   w1 = z1case - z1control ;

   z2case /= xcase ;
   z2control /= xcontrol ;

   w2 = z2case + z2control +1.0e-10  ;

   return w1/sqrt(w2) ;






}

double nick_score
 (Indiv *indiv_array, double **gams, int numindivs, 
 double xrisk, int chrom)
/** 
 cf: admix_score.tex 
 cases only used 
 assumes mult. risk model 
 psi(a) (admix_score.tex etc) = risk^a 
*/
{
  int *caseind, ncases, numstates = 3 ;
  int affstatus ;
  Indiv *indx ;
  double gg[3], psi[3], *tt, t2[3], t3[3] ;
  double totscore, score, s1, s2, theta ;
  int i, k, j ;

  ZALLOC(caseind,numindivs,int);

  psi[0] = 1 ;
  psi[1] = xrisk ;
  psi[2] = xrisk*xrisk ;


  ncases = 0 ;
  for (i=0; i<numindivs; i++) {
    indx = indiv_array + i ;  
    affstatus = indx -> affstatus;
    if (affstatus != 1) continue ;
    caseind[ncases] = i ;  
    ++ncases ;
  }  
 
  totscore = 0.0 ;
  for (k=0; k<ncases; ++k) { 
   i = caseind[k] ;
   indx = indiv_array + i ;  

   theta = indx->theta_mode ;
   if (chrom==23) theta = indx-> Xtheta_mode ;

   sett1(t2, theta, 2) ;
   sett1(t3, theta, 3) ;

   gg[0] = gams[0][i] ;
   gg[1] = gams[1][i] ;
   gg[2] = gams[2][i] ;

   if (asum(gg, 3) <= 0.0) continue ;
   tt = t3 ;
   numstates = 3 ;
   if ((chrom == 23) && ((indx -> gender) == 'M')) {
     tt = t2 ; numstates = 2 ;
   }
/** 
 this is the only place where male X stuff 
 is needed, except for settt 

 Note that for the Male X gg[2], tt[2]
 are both zero, so psi value is irrelevant 
*/

/* vdot is dot product */
   s1 = log10(vdot(psi, gg, numstates)) ;
   s2 = log10(vdot(psi, tt, numstates)) ;
   score = s1-s2 ;
   totscore += score ;

   if (debug) {  
    printf("score:  %9.3f  %9.3f\n", score, totscore) ;
   }
  }
  free(caseind) ;
  return totscore ;

}

double mapcheckx(SNP *cupt)
{
  Indiv *indx ;

  int markernum ;
  int chrom, numstates, indiv, ignore ;
  double gstar[3], X[3], tt[3] ;
  double ylike, ytotlike = 0.0 ;
  double s0, s1 ;
  double theta, lambda ;

  if (cupt -> isfake) return 0.0 ;
  if (cupt -> ignore) return 0.0 ;

  markernum = cupt -> markernum ;
  chrom = cupt -> chrom ;
  for (indiv = 0; indiv < numindivs; ++indiv) {
    indx = indiv_array + indiv ;
    gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ;

    if (ignore == YES) continue ;
    sett1(tt, theta, numstates) ;

    gammastar(gstar, markernum, indiv, numstates) ;
    if (gstar[0] < 0.0) continue ;
    tcalcX(cupt, indiv,  X, numstates) ;
    s0 = log10(vdot(gstar, X, numstates)) ;
    s1 = log10(vdot(tt, X, numstates)) ;
    ylike = s1-s0 ;
    ytotlike += ylike ;
  }

  return ytotlike ;

}

double freqcheckx(  SNP *cupt, double *rp, double *rpp, 
  int cmode, int vmode) 

/**
 cmode 0 controls 
 cmode 1 cases 
 cmode 2 all 
*/

{
  Indiv *indx ;

  int markernum ;
  int chrom, numstates, indiv ;
  int k, t ;
  int *obs ;
  double gstar[3], X[3], *tt ;
  double ylike, ytotlike = 0.0 ;
  double p, pp ;
  double **gg ;
  int *ismalex ;
  int affstatus ;

  markernum = cupt -> markernum ;
  *rp = *rpp = 0.5 ;
  if (cupt -> isfake) return 0.0 ;
  if (cupt -> ignore) return 0.0 ;

  ZALLOC(gg, 3, double *) ;
  for (k=0; k<3; k++) {             
   ZALLOC(gg[k], numindivs, double) ; 
  }
  ZALLOC(obs, numindivs, int) ;
  ZALLOC(ismalex, numindivs, int) ;

  chrom = cupt -> chrom ;
  for (indiv = 0; indiv < numindivs; ++indiv) {
    numstates = 3 ;
    ismalex[indiv] = NO ;
    indx = indiv_array + indiv ;
    if ((chrom == 23) && (indx -> gender == 'M')) {
      numstates = 2 ;
      gstar[2] = 0.0 ;
      ismalex[indiv] = YES ;
    }
    if (cmode<2) {
     indx = indiv_array + indiv ;  
     affstatus = indx -> affstatus;
     if (affstatus != cmode) ismalex[indiv] = -1  ;
    }
    gammastar(gstar, markernum, indiv, numstates) ;
    for (k=0; k<3; k++) {             
     gg[k][indiv]  = gstar[k] ; 
    }
    t = obs[indiv] = getgtypes(cupt, indiv)  ;
    if (vmode)  {  
     printf("ffv %4d %3d %9.3f %9.3f %9.3f\n", indiv, t,     
      gg[0][indiv], gg[1][indiv], gg[2][indiv]) ;
    }
  }

  p  = cupt -> adm_PopA_freq ;
  pp = cupt -> adm_PopB_freq ;

  ylike = reest(obs, gg,  numindivs, &p, &pp, ismalex) ;
/* 
 EM climb 
*/

  *rp = p ;
  *rpp = pp ;

  free(obs) ;
  free(ismalex) ;
  for (k=0; k<3; k++) {             
   free(gg[k]) ;
  }
  free(gg) ;

  return ylike ;

}
double
reest(int *obs, double **gg, 
 int numindiv, double *xp, double *xpp, int *ismalex)  
{
  double *X,  **XX ;
  double **Y, ***YY ;
  double **Z, ***ZZ ;
  double *W, **WW ; /* pointer to Y or Z */
  double p, pp, y, tt[3], post[3] ;
  double ylike, ybase, ylast, yinc = 999 ;
  double emgam[4] ;
  int k, n, i, a, numdata ;
  int numiter = 100, iter ;
  int numstates ; 
  Indiv *indx ;

  p = *xp ;
  pp = *xpp ;


  ZALLOC(XX, 4, double *) ;
  ZALLOC(X, 3, double)  ;
  for (k=0; k<4; k++)  
   ZALLOC(XX[k], 3, double) ;

  ZALLOC(YY, 3, double **) ;
  ZALLOC(ZZ, 3, double **) ;
  ZALLOC(Y, 3, double *) ;
  ZALLOC(Z, 3, double *) ;
  for (n=0; n<3 ; ++n) {
   ZALLOC(YY[n], 4, double *) ;
   ZALLOC(ZZ[n], 4, double *) ;
   ZALLOC(Y[n], 3, double)  ;
   ZALLOC(Z[n], 2, double)  ;
   for (k=0; k<4; k++)  {
    ZALLOC(YY[n][k], 3, double) ;
    ZALLOC(ZZ[n][k], 2, double) ;
   }
  }

  for (iter= 1; iter <= numiter ; ++iter) {  
   numdata = 0 ;
   for (n=0; n<3; ++n) {
    docalcXX(n, Y[n], YY[n], p, pp, 3) ;
    docalcXX(n, Z[n], ZZ[n], p, pp, 2) ;
   }
   ylike = 0.0 ;
   vzero(emgam, 4) ;
   for (i=0; i<numindiv; i++) {   
     if (ismalex[i] < 0 ) continue ;
     n = obs[i] ;
     if (n<0) continue ;
     indx = indiv_array + i ;
     if (indx -> ignore) continue ;
     W = Y[n]  ;
     WW = YY[n]  ;
     numstates = 3 ;
     if (ismalex[i] == YES) {
       numstates = 2 ;
       W = Z[n]  ;
       WW = ZZ[n]  ;
     }
     for (a=0; a<numstates; a++) {
      tt[a] = gg[a][i] ;
     }
     if (tt[0]<0.0) continue ;
     ++numdata ;
     vvt(post, W, tt, numstates) ;
     y = asum(post, numstates) ;
     ylike += log(y) ;
     vst(post, post, 1.0/y, numstates) ;
     for (k=0; k<4; k++) { 
      emgam[k] += vdot(WW[k], post, numstates) ;
     }
   }
   if (iter==1) ybase = ylike ;

   ylike -= ybase ;

   if (iter>1) yinc = ylike - ylast ;

/**
   printf("iter:  %3d  like:  %9.3f %9.3f %9.3f\n",iter, ylike,p, pp) ;
*/

   if (numdata == 0) break ;

   y = emgam[0]+emgam[1] ;
   p = emgam[0]/y ;
   y = emgam[2]+emgam[3] ;
   pp = emgam[2]/y ;

   if (yinc< 0.01) break ;
   ylast = ylike ;
  }

  for (k=0; k<4; k++)  
   free(XX[k]) ;
  free(XX) ;
  free(X) ;

  for (n=0; n<3 ; ++n) {
   for (k=0; k<4; k++)  {
    free(YY[n][k]) ;
    free(ZZ[n][k]) ;
   }
   free(YY[n]) ;
   free(ZZ[n]) ;
   free(Y[n]) ;
   free(Z[n]) ;
  }

  free(YY) ;
  free(ZZ) ;
  free(Y) ;
  free(Z) ;

  *xp = p ;
  *xpp = pp ;

  return ylike ;
  
}

void docalcXX(int x, double *X, double **XX,
 double p, double pp, int numstates) 
{
  int k ; 
  double A[3], B[3], C[3], D[3], T[3] ;
  double q, qq ;

  vzero(A,3) ;
  vzero(B,3) ;
  vzero(C,3) ;
  vzero(D,3) ;

  if (x<0) {  
   vclear(X, 1.0, numstates) ;
   copyarr(A, XX[0], numstates) ;
   copyarr(B, XX[1], numstates) ;
   copyarr(C, XX[2], numstates) ;
   copyarr(D, XX[3], numstates) ;
   return ; 
  }

  if ((p<0) || (pp<0)) 
    fatalx("marker freqs %s not set\n") ;

  q = 1.0-p ;
  qq = 1.0-pp ;

  if (numstates == 3) {
    switch(x) {
      
    case 2:
      X[0] = p * p;
      X[1] = p * pp;
      X[2] = pp * pp;

      A[0] = 2 * p * p ;
      A[1] = p * pp ;

      C[1] = p * pp ;
      C[2] = 2 * pp * pp ;  
      break;
      
    case 1:
      X[0] = 2 * p * q ;
      X[1] = pp * q + qq * p ;
      X[2] =  2 * pp * qq ;

      A[0] = 2 * p * q ;
      A[1] = qq * p ;     

      B[0] = 2 * p * q ;
      B[1] = pp * q ;  

      C[1] = pp * q ; 
      C[2] =  2 * pp * qq ; 

      D[1] = qq * p ;
      D[2] = 2 * pp *qq ;
      break;
      
    case 0:
      X[0] = q * q;
      X[1] = q * qq;
      X[2] = qq * qq;

      B[0] = 2 * q * q ;
      B[1] = q * qq ;

      D[1] = q * qq ;
      D[2] = 2 * qq * qq ;  
      break;
  
    default:
      fatalx(" genotype unset\n") ; 
      break;
    }
  }
  else if (numstates == 2) {
    switch(x) {

    case 1:
    case 2:
      A[0] = X[0] = p;
      C[1] = X[1] = pp;
      break;

    case 0:
      B[0] = X[0] = q;
      D[1] = X[1] = qq;
      break;
  
    default:
      fatalx(" genotype unset\n") ; 
      break;
    }
  }
   vsp(T, X, 1.0e-6, numstates) ;
   vvd(XX[0], A, T, numstates) ;
   vvd(XX[1], B, T, numstates) ;
   vvd(XX[2], C, T, numstates) ;
   vvd(XX[3], D, T, numstates) ;
   return ;
}

void doleave1(double *topscore, double *botscore, 
  SNP  *cupt, int lrmode, int numsnps, int winwidth) 
{

    Indiv *indx ;
    double aa[3], bb[3], ww[3], y ;
    double theta, lambda ;
    int markernum ; 
    int numstates ;
    int wlen, i, j, k, indiv ;
    int left, right, nfake ;
    int chrom, fc, lc ;
    double **gammas ;
    double ***wingams ;
    double ysc ;
    double *apt, *bpt ;

     markernum = cupt -> markernum ;
     
    /* wingams[indiv][0..2][markernum] */
    if (verbose) 
     printf("leave1 marker %d\n",markernum) ;
    
    ZALLOC(wingams, numindivs, double **) ;
    for (i=0; i<numindivs; i++) {  
     ZALLOC(wingams[i], 3, double *) ;
    }
    ZALLOC(gammas, 3, double *) ;
    for (k=0; k<3; k++) {  
     ZALLOC(gammas[k], numindivs, double ) ;
    }
    if (cupt -> isfake) return ;
    if (cupt -> ignore) return ;
    chrom = cupt -> chrom ; 
    
    fc = fchrom[chrom] ;
    lc = lchrom[chrom] ;
    nfake = 0 ;  
    if (lrmode == 0) {
     left = markernum ;
     right = lc ;
     for (i=markernum+1 ; i <= lc; i++) { 
      if ( !(snpmarkers[i] -> isfake) ) continue ;
      ++nfake ;
      right = i ; 
      if (nfake >= winwidth) {       
        break ;
      }
     }
    }
    else  {
     right = markernum ;
     left = fc ;
     for (i=markernum-1 ; i >= fc; i--) { 
      if ( !(snpmarkers[i] -> isfake) ) continue ;
      ++nfake ;
      left = i ;
      if (nfake >= winwidth) {       
        break ;
      }
     }
    }
    left =  nextfake(left, fc, -1) ; 
    right = nextfake(right, lc, 1) ; 
/* try and make boundaries fake markers */
    wlen = right - left + 1 ;
    for (indiv = 0; indiv < numindivs; ++indiv) { 
     for (k=0; k<3; k++) {
      ZALLOC(wingams[indiv][k], wlen, double) ;
     }
    }
    for (indiv = 0; indiv < numindivs; ++indiv) { 
     indx = indiv_array + indiv ;
     gettln(cupt, indx, &theta, &lambda, &numstates, NULL) ;

    if (lrmode==0) {
     alphastar(aa, left, indiv, numstates) ;
    }

    else {
     grabgam(aa, left, indiv, 0) ; /* alpha */
    }

/** 
 it is correct but too tricky to always get alphastar
 on left as when lrmode = 1 left is fake
*/

    grabgam(bb, right, indiv, 1) ; /* beta */
    apt = aa ; 
    bpt = bb ; 
    if (apt[0] < 0.0) apt = NULL ;
    if (bpt[0] < 0.0) bpt = NULL ;

    HMMwindow ( snpmarkers, indiv, theta, lambda, left, right, markernum,
     wingams[indiv], apt, bpt, numstates) ;
   }

   for (j=0; j<wlen; j++) {  
     copywingams(wingams, gammas, numindivs, j) ;
     ysc = nick_score(indiv_array, gammas, numindivs, risk, chrom);
     if ((left+j)>=numsnps) {
       printf("%d %d %d %d\n",markernum, left, right, lc) ;
       fatalx("%d bnds1\n", markernum+j) ;
     }
     if ((left+j) == markerpub) printf("markerpub: %d %d %d %9.3f\n", 
      markernum, left, right, ysc) ;
     topscore[left+j] = MAX(topscore[left+j], ysc) ;
     botscore[left+j] = MIN(botscore[left+j], ysc) ;
   }
    
   
/* destructor */

    for (k=0; k<3; k++) {
      free(gammas[k]) ;
    }
    free(gammas) ;

    for (indiv = 0; indiv < numindivs; ++indiv) { 
     for (k=0; k<3; k++) {
      free(wingams[indiv][k]) ;
     }
     free(wingams[indiv]) ;
    }
    free(wingams) ;
}

void
copywingams(double ***wingams, double **gammas, int n, int col)  
/* copy one column of wingams */
{
  int k, i ;
  for (k=0; k<3; k++) 
   for (i=0; i<n; i++)  
     gammas[k][i] = wingams[i][k][col] ;
}


int nextfake (int start, int end, int inc) 
/* search for fake marker */
{
  int x ;
  x = start ;
  if ( (x<end) && (inc <0)) 
    fatalx("bad nextfake %d %d %d\n",start, end, inc ) ;
  if ( (x>end) && (inc >0)) 
    fatalx("bad nextfake %d %d %d\n",start, end, inc ) ;
  for (x=start; x != end ; x += inc)  {
   if (snpmarkers[x] -> isfake)  return x ;
  }
  return end ;
}


void pubscores (double *basescore, double *topscore, double *botscore, 
 SNP **snpmarkers, int numsnps) 
{
   int i, nmark=0 ;
   int chrom ;
   SNP *cupt ;

   double be[NCHROMP], tb[NCHROMP], bb[NCHROMP] ;
   vclear(be, -100.0, NCHROMP) ;
   vclear(tb, -100.0, NCHROMP) ;
   vclear(bb, -100.0, NCHROMP) ;
   printf("\n\n scores for each fake\n") ;
   printf("##base: Score for the marker without using leave1out, min and max are the minimum and maximum scores obtained when using the leave1out algorithm\n");
   printf("%5s %20s %5s", "SNP_Index", "SNP_ID","chrom") ;
   printf(" %9s %9s %9s\n", "base", "min", "max" ) ;
   for (i=0; i<numsnps; i++) {
    cupt = snpmarkers[i] ;  
    chrom = cupt -> chrom ;
    if (!(cupt -> isfake)) continue ;
    printf("%5d %3d %20s", i, chrom, cupt->ID) ;
    printf(" %9.3f %9.3f %9.3f\n", 
       basescore[i], botscore[i], topscore[i]) ;
    be[chrom] = MAX(be[chrom], basescore[i]) ;
    bb[chrom] = MAX(bb[chrom], botscore[i]) ;
    tb[chrom] = MAX(tb[chrom], topscore[i]) ;
   }
   
  printf("\n\n") ;
   printf("%20s %5s %9s %9s %9s\n", 
    "", "chrom", "base", "min", "max") ; 
   for (chrom=1; chrom<= NCHROM; ++chrom) {  
     if (be[chrom] < -50.0) continue ;
     printf("%20s %5d %9.3f %9.3f %9.3f", 
       "best score (chrom)",chrom, be[chrom], bb[chrom], tb[chrom]) ;
     printf("\n") ;
   }

}
/**** execute HMM on a window ****/
double HMMwindow
 (
 SNP ** snpmarkers, 
 int indiv,
 double theta, 
 double lambda, 
 int first, 
 int last, 
 int leave1out,  /* -1 if none */
 double **gammas,
 double *ainit, 
 double *binit, 
 int numstates )
/* returns log10 score */

/**
 indiv is individual number 
 theta, lambda obvious  (note must be set on input)
 first, last snpmarker numbers
 leave1out marker number to ignore data for
 gammas[k][c] is post prob for k c alleles marker first+c
 ainit , binit alpha beta initializers.  Can be NULL
 numstates 2 or 3.
*/

{

  SNP *cupt ;
  SNP tempsnpm ;
  int j,k,x,y,count ;

  double **alpha, **beta ;
  double loglike ;
  double ta[3], tb[3], yy ; 
  int nummarkers ;
  int numsta = 3 ;
  int dowork = YES ;  /* no we just skip */

  double *aaa, *bbb ;

  static int ncall = 0 ;
  int dbug ;

  count = nummarkers = last-first+1 ;

/** 
 the line below is perhaps doubtful
 and not compatible with dumpgams  
 we leave it in for now.
*/

  if ( (theta<.0001) || (theta>.9999) || (lambda<.0001)) dowork = NO ;
  if ((ainit != NULL) && (ainit[0] < 0.0))   dowork = NO ;
  if ((binit != NULL) && (binit[0] < 0.0))   dowork = NO ;

  if (dowork == NO)        {
   for (k=0; k<numstates; k++)  {  
    vclear(gammas[k], -1.0, count) ;
   }
   return 0.0 ;
  }

  ++ncall ;
  
/** 
 slighly dubious code
 we overwrite an important array (snpmarkers) 
 and fix on exit
*/
  if (leave1out>=0) {   
   cupt = snpmarkers[leave1out] ;
   tempsnpm = *cupt ;
   cupt -> isfake = YES ;
  }

  alpha = initarray_2Ddouble(nummarkers,numsta,0.0);
  beta = initarray_2Ddouble(nummarkers,numsta,0.0);

  HMMbeta(snpmarkers,indiv, theta, lambda, first, last,
	beta, binit, numstates, NULL) ;

  loglike = HMMalpha(snpmarkers,indiv, theta, lambda, first, last,
	alpha, ainit, numstates, NULL) ;


  for (k=0; k<numstates; k++)  {  
   vzero(gammas[k],  count) ;
  }

  /* ############### Posterior Probabilities ################# */

/** 
 ###########################################
 dangerous bend.  gammas based at zero NOT first 
 Change from Neil's code 

 NJP
*/

  for (j=first; j<=last; ++j) {
    count = j-first ;
    aaa = alpha[count] ;
    bbb = beta[count] ;
    vvt(ta, aaa, bbb, numstates) ;  
/* Vector Vector Times */
    yy = asum(ta, numstates) ;
    vst(ta, ta, 1.0/yy, numstates) ; 
    for (k=0; k<numstates; k++) {
     gammas[k][count] = ta[k] ;
    }
  }

  free2D(&alpha, nummarkers);
  free2D(&beta, nummarkers);

  if (leave1out>=0) {  
   cupt = snpmarkers[leave1out] ;
   *cupt  = tempsnpm;
  }


  return loglike ;

}
void dophyscheck(SNP **snpm, int numsnps) 
{
// catch places where physpos genpos are in opposite order
  SNP *cupt, *cuptold ;
  int i ;

  printf("###PHYSCHECK RESULTS FOLLOW:\n");
  printf("##%15s %14s %14s %14s %14s %14s\n","SNP1_ID","SNP2_ID","SNP1_GEN_POS","SNP2_GEN_POS","SNP1_PHYS_POS","SNP2_PHYS_POS");
  for (i=0; i<numsnps; i++) {   
   cupt = snpm[i] ;
   if (i==0) cuptold = cupt ;
   if (cupt -> isfake) continue ;
   if (cupt -> chrom == cuptold -> chrom)  {
    if (cupt -> physpos < cuptold -> physpos) {  
     printf("physcheck %20s %15s %12.3f %12.3f %13.0f %13.0f\n", 
     cuptold->ID, cupt -> ID, 
     cuptold -> genpos, cupt -> genpos, 
     cuptold -> physpos, cupt -> physpos);
    }
   }
   cuptold = cupt ;
  }
}

void checkindiv (SNP **snpm, Indiv **indm, int numsnps, int numindivs) 
// look at deviations from expected log -likeihood distribution 
// test designed after discussion with DR.  NJP 
{

  SNP *cupt;
  Indiv *indx ;

  int indiv, i ;
  int n, g ;
  int nchi ;  
  double xx[3], xxl[3] ;

  double theta, p, pp, f ;
  double lm, lm2, lchi, ldev, lsig, lvar ;
  double totchi, totdev, totsig, totvar ;

  printf("\n###CHECKINDIV RESULTS\n");
  printf("##%27s %11s %21s\n","Indiv_ID","theta","Z-score");
  for (indiv = 0; indiv < numindivs; ++indiv) {  
    indx = indm[indiv] ;  
    if (indx -> ignore) continue ;
    theta = indx -> theta_mode ;
    totdev = totchi = totvar = 0.0 ;
    nchi = 0 ;
    for (i=0; i<numsnps; i++) {  
     cupt = snpm[i] ;  
     if (cupt -> isfake) continue ;
     if (cupt -> ignore) continue ;
     if (cupt -> chrom > 22) continue ;  // autosomes only
     n = cupt -> ngtypes ;
     if (n==0) continue ;
     g = getgtypes(cupt, indiv) ; 
     if (g<0) continue ;

     p  = cupt -> adm_PopA_freq ;
     pp = cupt -> adm_PopB_freq ;
     f = theta*pp + (1.0-theta)*p ; /* expected allele freq */
     sett1(xx, f, 3) ;   /* prob dis of g */  
     vlog(xxl, xx, 3) ;
     lm = vdot(xx, xxl, 3) ;
     ldev = xxl[g] - lm ;
     vvt(xxl, xxl, xxl, 3) ;
     lm2 = vdot(xx, xxl, 3) ;  /* second moment about 0 */
     lvar = lm2 - lm*lm ;
     lsig = ldev/sqrt(lvar) ;
     lchi = lsig*lsig ;
     totdev += ldev ;
     totchi += lchi  ;
     totvar += lvar ;
     ++nchi ;
     if ((indiv < -1) && (i<10)) {  
      printf("zz%2d %2d %9.3f %9.3f %9.3f %9.3f %9.3f %2d", indiv, i, f, ldev, xx[0], xx[1], xx[2], g) ;
      printf(" %9.3f %9.3f %9.3f %9.3f\n", lchi, lvar, totdev, totchi) ;
     }
    }
//stats
    if (nchi==0) { 
      printf("*** warning no autosomal data for indiv: %s\n", indx -> ID) ;
      continue ;
    }

    printf("checkindiv %20s %9.3f", indx -> ID, theta) ;
    printf(" %9.3f", totdev) ;
    lsig = totdev/sqrt(totvar) ;
    printf(" %9.3f", lsig) ;  // main statistic??  
    printf(" %4d  %9.3f", nchi, totchi) ;
    totchi /= (double) nchi ;
    printf("%9.3f",  totchi) ;
    printf("\n") ;
  }
}

void corrgamma (SNP **snpm,  int numsnps, int numindivs, 
  double winlen, double thresh, double epith) 
/* looks at correlation of probit(gamma - theta) */
{
    int i, j, k, chrom, n1, n2 ; 
    SNP *cupt, *cupt2 ; 
    double *v1, *v2, dis, y, sig, *q1, *q2 ;
    static int nhit = 0 ;

   ZALLOC(v1, numindivs, double) ;
   ZALLOC(v2, numindivs, double) ;

   ZALLOC(q1, numindivs, double) ;
   ZALLOC(q2, numindivs, double) ;

    for (i=0; i< numsnps; i++)  { 
      cupt = snpm[i] ;  
      chrom = cupt -> chrom ;
      if (isitpeak(snpm, i, winlen, thresh))  {
        printf("peak: %20s %3d %13.0f %9.3f\n", 
        cupt->ID, chrom, cupt -> physpos, cupt -> score) ;
        grabvx(cupt, v1, &n1) ;
        if (n1 == 0) continue ;
        probit(q1, v1, n1) ;

        for (j=0; j<numsnps; j++) {  
          cupt2 = snpm[j] ;  
          if (cupt2 -> isfake != YES) continue ;
          dis = cupt -> genpos - cupt2 -> genpos ;
          if (cupt -> chrom != cupt2 -> chrom) dis = 99999.0 ;  
          if (fabs(dis) <= winlen) continue ;
           grabvx(cupt2, v2, &n2) ;
           if (n2 == 0) continue ;
           if (n1 != n2) fatalx("bad grabx %d %d\n", n1, n2) ;
           probit(q2, v2, n1) ;
           y = corr(q1, q2, n1) ;  
           sig = sqrt( (double) n1) * y ;  
           if (fabs(sig) > epith) { 
            ++nhit ;
            printf("corrgamm %3d %3d %13.0f %13.0f %9.3f %9.3f\n", 
             cupt -> chrom, cupt2 ->chrom, cupt -> physpos, cupt2 -> physpos, sig, y) ;
           }
        }
      }
    }

    free(v1) ; 
    free(v2) ;
    free(q1) ; 
    free(q2) ;

}
void
grabvx(SNP *cupt, double *vv, int *np) 
{
   int i, n=0, markernum ; 
   Indiv *indx ;  
   double gg[3], gave  ;

   markernum = cupt -> markernum ;
   for (i=0; i< numindivs; i++)  {  
     indx = indivmarkers[i] ;  
     if (indx -> affstatus != YES) continue ;  
     grabgam(gg, markernum, i, 2)  ;  
     if (gg[0] < 0.0) gave = 0.0 ; 
     else {  
      gave = 0.5*(gg[1] + 2.0*gg[2]) ;
      gave -= indx -> theta_mode ;
     }
     vv[n] = gave ; 
     ++n ;
   }

  *np = n ;

}

static int 
isitpeak(SNP **snpm,  int markernum, double winlen, double thresh) 
{

   int i, c1, fc, lc, lo, hi, k ;  
   double s1, s2, dis ;
   SNP *cupt, *cupt2 ;
   static int ncall=0  ;  


   ++ncall ; 
// if (ncall > 100) return NO ;
   

   
   k = markernum ;
   cupt = snpm[k] ;
   if (!(cupt -> isfake)) return NO ;

   s1 = cupt -> score ;

/**
    if (s1  < thresh) return NO ;
*/

   c1 = cupt -> chrom ;
   fc = fchrom[c1] ;
   lc = lchrom[c1] ;
   lo = MAX(fc, k-20) ; 
   hi = MIN(lc, k+20) ; 
   for (i= lo; i<=hi; i++) {  
    if (i==k) continue ;
    cupt2 = snpm[i] ;  
    if (!(cupt2 -> isfake)) continue  ;
    dis = cupt -> genpos - cupt2 -> genpos ; 
//  printf("zzz %9.3f  %9.3f\n", dis, winlen) ;
    if (fabs(dis) > winlen) continue ;
     s2 = cupt2 -> score ;  
     if (s2>s1) return NO ;
   }
   return YES  ;
}
