#include <fcntl.h>
#include <ctype.h>
#include  "mcio.h" 

int  usecm = NO ;
int plinkinputmode = NO ;
static int debug = NO ;
extern  char *trashdir ;
extern int qtmode ;
extern int verbose ;
extern int familynames ;
extern double lp1, lp2  ;
extern double a1, b1  ;

extern int packmode ;  
extern char *packgenos ;
extern char *packepath ;
extern int packlen, rlen ;
extern int malexhet ;

int tempnum = 0 ;
int tempfake = 0 ;

static int *pedcols = NULL ;

static int maxgenolinelength = -1 ;
static int tersemode = NO ;
enum outputmodetype outputmode = ANCESTRYMAP ;
static double maxgpos[MAXCH] ;

int getbedgenos(char *gname, SNP **snpmarkers, Indiv **indivmarkers, 
 int numsnps, int numindivs, int nignore)  ;

void freeped()  ;

static char x2base(int x) ;
static void gtox(int g, char *cvals, int *p1, int *p2)  ;

int ancval(int x)  ;
void printalleles(SNP *cupt, FILE *fff) ;

int getsnps(char *snpfname, SNP ***snpmarkpt, double spacing,
  char *badsnpname, int *numignore, int numrisks) 
// returns number of SNPS
// numrisks 
{
/* read file of real SNPS store in temporary structure */

 SNPDATA **snpraw, *sdpt ;
 static SNP **snpmarkers ;
 SNP *cupt ;
 int **snppos ;
 int nreal, nfake, numsnps = 0, i ;
 int *snpindx ;
 double xspace ;

 xspace = spacing ;
 nreal = getsizex(snpfname) ;
 if (nreal <= 0) fatalx("no snps found: snpfname: %s\n", snpfname) ;
 ZALLOC(snpraw, nreal, SNPDATA *) ;
 if (pedcols == NULL) {
  ZALLOC(pedcols, nreal, int) ;  
  ivclear(pedcols, -1, nreal) ;
 }
 for (i=0; i<nreal ; i++) { 
  ZALLOC(snpraw[i], 1, SNPDATA) ;
  snpraw[i] -> inputrow = -1 ;
 }
 readsnpdata(snpraw, snpfname) ;
 dobadsnps(snpraw, nreal, badsnpname) ;

 ZALLOC(snppos, nreal, int *) ;
 for (i=0; i<nreal; i++) {  
  ZALLOC(snppos[i], 3, int) ;
 }


 for (i=0; i<nreal ; i++) {
  sdpt = snpraw[i] ;
  snppos[i][0] = sdpt -> chrom ;   
  if (sdpt->ignore) snppos[i][0] = 99 ;
  snppos[i][1] = nnint((sdpt -> gpos)*GDISMUL) ;       
  snppos[i][2] = nnint(sdpt -> ppos) ;      
 }

 ZALLOC(snpindx, nreal, int) ;
 ipsortit(snppos, snpindx, nreal, 3) ;

 if ((usecm) && (xspace>0.5)) { 
   printf("*** warning fake spacing given in cM\n") ;
   xspace /= 100.0 ;
 }
 
// get number of fakes

 nfake = numfakes(snpraw, snpindx, nreal, xspace) ;
 numsnps = nreal + nfake ;
 
 tempnum = numsnps ;
 tempfake = nfake ;

// allocate storage

 ZALLOC(snpmarkers, numsnps, SNP *) ;
 for (i=0; i<numsnps; i++) {  
  ZALLOC(snpmarkers[i], 1, SNP) ;
  cupt = snpmarkers[i] ;
  clearsnp(cupt) ;
  ZALLOC(cupt -> modelscores, numrisks, double) ;
  ZALLOC(cupt -> totmodelscores, numrisks, double) ;
 }
 *snpmarkpt = snpmarkers ;
  numsnps = loadsnps(snpmarkers, snpraw, snpindx, nreal, xspace, numignore) ;


// and free up temporary storage
 for (i=0; i<nreal ; i++) { 
  free(snpraw[i]) ;             
  free(snppos[i]) ;             
 }
 free(snpraw) ;
 free(snppos) ;
 free(snpindx);

 /* printf("numsnps: %d\n", numsnps) ; */

/**
 if (pedcols != NULL) { 
  printimat(pedcols, 1, MIN(100, numsnps)) ;
 }
*/

 return numsnps ;
}


int getsizex(char *fname) 
// number of real lines 
{
 char line[MAXSTR+1], c ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0 ;
 int skipit ;
 int len ;

 FILE *fff ;
 openit(fname, &fff, "r") ;
 line[MAXSTR] = '\0' ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') skipit = YES ;
   if (strcmp(sx,"SNP_ID") == 0) skipit = YES ;
   if (strcmp(sx,"Indiv_ID") == 0) skipit = YES ;
   if (skipit == NO) {
    ++num ;
   }
// now flush the rest of the line if necessary.
   len = strlen(line) ;  
   c = line[len-1] ;
   if (c != '\n') { 
     while ((c = fgetc(fff)) != EOF) {  
      if (c == '\n') break ;
     }
   }
   freeup(spt, nsplit) ;
   continue ;
 }
 fclose(fff) ;
 fflush(stdout) ;
 return num ;
}

int ismapfile(char *fname) 
// PLINK map file ? 
// just look at file name (perhaps should look at format) 
{

 char *sx ;
 int len ;
 len = strlen(fname) ;
 if (len<4) return NO ;
 sx = fname+len-4 ;

 if (strcmp(sx, ".map") == 0) return YES  ;
 if (strcmp(sx, ".bim") == 0) return YES  ;
 
 if (len<7) return NO ;
 sx = fname+len-7 ;
 if (strcmp(sx, ".pedsnp") == 0) return YES  ;

 return NO ;

}
int ispedfile(char *fname) 
// PLINK ped file ? 
// just look at file name (perhaps should look at format) 
{

 char *sx ;
 int len ;
 len = strlen(fname) ;
 if (len<4) return NO ;
 sx = fname+len-4 ;

 if (strcmp(sx, ".ped") == 0) return YES  ;
 if (strcmp(sx, ".fam") == 0) return YES  ;

 if (len<7) return NO ;
 sx = fname+len-7 ;
 if (strcmp(sx, ".pedind") == 0) return YES  ;

 return NO ;

}

int isbedfile(char *fname) 
// PLINK ped file ? 
// just look at file name (perhaps should look at format) 
{

 char *sx ;
 int len ;
 len = strlen(fname) ;
 if (len<4) return NO ;
 sx = fname+len-4 ;

 if (strcmp(sx, ".bed") == 0) return YES  ;
 return NO ;

}

int readsnpdata(SNPDATA **snpraw, char *fname) 
{
 char line[LONGSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0, k ;
 int skipit ;
 SNPDATA *sdpt ;

 FILE *fff ;
 int chrom ;

 if (ismapfile (fname)) {  
    plinkinputmode = YES ;
    return readsnpmapdata(snpraw, fname)  ;
 }
 

 vclear(maxgpos, -9999.0, MAXCH) ; 
 openit(fname, &fff, "r") ;
 while (fgets(line, LONGSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') skipit = YES ;
   if (strcmp(sx,"SNP_ID") == 0) skipit = YES ;
   if (strcmp(sx,"Indiv_ID") == 0) skipit = YES ;
   if (skipit == NO) { 
    if (nsplit<4) fatalx("(readsnpdata) bad line: %s\n",line) ;
    sdpt = snpraw[num] ;
    sdpt -> inputrow = num ;
    if (strlen(spt[0]) >= IDSIZE) fatalx("ID too long\n", spt[0]) ;
    strcpy(sdpt->ID, spt[0]) ;

    if (spt[1][0] == 'X') {
     sdpt -> chrom = 23 ;
    }

    else {
     sdpt -> chrom = atoi(spt[1]) ;
    }
    chrom = sdpt -> chrom ;
    if (chrom>=MAXCH) fatalx("bad chrom: %s\n", line) ;
    if (chrom==0) fatalx("bad chrom: %s\n", line) ;

    sdpt -> gpos = atof(spt[2]) ;
    if (sdpt->gpos > 100) {
      if (sdpt->gpos > 1.0e6) 
        fatalx("absurd genetic distance:\n%s\n", line) ;
     if (!usecm) {
      printf("*** warning.  genetic distances are in cM not Morgans\n") ;
      printf("%s\n",line) ;
     }
     usecm = YES ;
    }
    maxgpos[chrom] = MAX(maxgpos[chrom], sdpt -> gpos) ;
    sdpt -> ppos = atof(spt[3]) ;
    if (nsplit<8) { 
     ivzero(sdpt->nn,4) ; 
     if (nsplit==6) {  
      sx = spt[4] ; sdpt -> alleles[0] = sx[0] ;
      sx = spt[5] ; sdpt -> alleles[1] = sx[0] ;
     }
    }
    else {
     for (k=0; k<4; k++) { 
      sdpt->nn[k] = atoi(spt[4+k]) ;
     }
     if (nsplit==10) {  
      sx = spt[8] ; sdpt -> alleles[0] = sx[0] ;
      sx = spt[9] ; sdpt -> alleles[1] = sx[0] ;
     }
    }
    ++num ;
   }
   freeup(spt, nsplit) ;
   continue ;
 }
 if (usecm) {
  for (k=0; k<num ; ++k) {
   snpraw[k] -> gpos /= 100.0 ;
  }
 }
 
 fclose(fff) ;
 return num ;
}

int readsnpmapdata(SNPDATA **snpraw, char *fname) 
{
 char line[MAXSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0, k ;
 int skipit, len, maxg=0 ;
 SNPDATA *sdpt ;
 int numbadc = 0 ; 

 FILE *fff ;
 int chrom ;


 vclear(maxgpos, -9999.0, MAXCH) ; 
 openit(fname, &fff, "r") ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') skipit = YES ;
   if (skipit == NO) { 
    if (nsplit<4) fatalx("(readsnpmapdata) bad line: %s\n",line) ;
    sdpt = snpraw[num] ;
    if (strlen(spt[1]) >= IDSIZE) fatalx("ID too long\n", spt[1]) ;
    strcpy(sdpt->ID, spt[1]) ;
   

    if (nsplit >=6 ) {  
     sx = spt[4] ; sdpt -> alleles[0] = sx[0] ;
     sx = spt[5] ; sdpt -> alleles[1] = sx[0] ;
    }
    else { 
     cclear((unsigned char *) sdpt -> alleles, CNULL,  2) ;
    }

    if (sdpt->alleles[0] == '0') sdpt -> alleles[0] = 'X' ; 
    if (sdpt->alleles[1] == '0') sdpt -> alleles[1] = 'X' ; 

    sx = spt[0] ;
    sdpt -> chrom = ptoachrom(sx) ;
    chrom = sdpt -> chrom ;

    if ((chrom==0) || (chrom>=MAXCH)) {   
      if (numbadc==0) printf("*** warning chromosome %d not supported\n") ; 
      ++numbadc ; 
      freeup(spt, nsplit) ; 
      continue ; 
    } 


    sdpt -> gpos = atof(spt[2]) ;
    if (sdpt->gpos > 100) {
      if (sdpt->gpos > 1.0e6) 
        fatalx("absurd genetic distance:\n%s\n", line) ;
     if (!usecm) {
      printf("*** warning.  genetic distances are in cM not Morgans\n") ;
      printf("%s\n",line) ;
     }
     usecm = YES ;
    }
    maxgpos[chrom] = MAX(maxgpos[chrom], sdpt -> gpos) ;
    maxg = MAX(maxg, maxgpos[chrom]) ;
    sdpt -> ppos = atof(spt[3]) ;
    if (nsplit<8) { 
     ivzero(sdpt->nn,4) ; 
    }
    else {
     for (k=0; k<4; k++) { 
      sdpt->nn[k] = atoi(spt[4+k]) ;
     }
    }
    sdpt -> inputrow = num ;
    ++num ;
   }
   freeup(spt, nsplit) ;
   continue ;
 }
 if (maxg==0) {
  usecm = NO ;
  for (k=0; k<num ; ++k) {
   snpraw[k] -> gpos = 1.0e-8 * snpraw[k] -> ppos ;
  }
 }
 if (usecm) {
  for (k=0; k<num ; ++k) {
   snpraw[k] -> gpos /= 100.0 ;
  }
 }
 if (pedcols == NULL) {
  ZALLOC(pedcols, num, int) ;  
  ivclear(pedcols, -1, num) ;
 }
 
 fclose(fff) ;
 return num ;
}


int
numfakes(SNPDATA **snpraw, int *snpindx, int nreal, double spacing) 

// it seems better for this internal routine 
// to use the precomputed values
{
  int nignore, numsnps;
  int nfake = 0, i, k, indx ;
  int num=0; 
  SNP *cupt ;
  SNPDATA *sdpt ;
  char *sname ;
  int *sp ;
  int xc = 0, chrom ;
  double fakedis, realdis ;   // gpos for fake marker 
  double physpos ;

  if (spacing <= 0.0) fakedis = 1.0e20 ;

  for (k=0; k< nreal ; k++) {  

   indx = snpindx[k] ;
   sdpt = snpraw[indx] ;
   if (sdpt->ignore) continue ;

   chrom = sdpt -> chrom ;
   realdis = sdpt -> gpos ;           

   if (chrom != xc)  {  
    fakedis = nextmesh(realdis, spacing) ;
    xc = chrom ;
   }
   while (fakedis<realdis) {  
     fakedis += spacing ;
     ++nfake ;
   }
  }
   return nfake ;
}
double nextmesh(double val, double spacing)  
{
 if (spacing==0.0) return 1.0e8 ;
 return ceil(val/spacing)*spacing ;
}


int
loadsnps(SNP **snpm, SNPDATA **snpraw, 
  int *snpindx, int nreal, double spacing, int *numignore) 

// snppos, snpindx could be recalculated but 
// it seems better for this internal routine 
// to use the precomputed values
// do NOT call externally
{
  int nignore, numsnps;
  int nfake = 0, i, k, indx ;
  int num=0; 
  SNP *cupt ;
  SNPDATA *sdpt ;
  char *sname ;
  int *sp ;
  int xc = 0, chrom ;
  double fakedis, realdis, xrealdis ;   // gpos for fake marker 
  double physpos ;
  double xl, xr, xmid, al, ar, fraw ;
  int nn[2], n0, n1 ;
  int cnum ;
  int inputrow ;


  if (spacing <= 0.0) fakedis = 1.0e20 ;
  for (k=0; k< nreal ; k++) {  


   indx = snpindx[k] ;
   sdpt = snpraw[indx] ;
   if (sdpt->ignore) continue ;

   chrom = sdpt -> chrom ;
   sname = sdpt -> ID ;
   realdis = sdpt -> gpos ;           
   physpos = sdpt -> ppos ;
   inputrow = sdpt -> inputrow ;

   if (chrom != xc)  {  
    fakedis = nextmesh(realdis, spacing) ;
    xc = chrom ;
    cnum = 0 ;
   }

   while (fakedis<realdis) {  
     if (cnum==0) break ;
     if (nfake>=tempfake)  fatalx(" too many fake markers (bug) %d %d\n", num, nfake) ;
     if (num>=tempnum)  fatalx(" too many markers (bug) %d %d\n", num, nfake) ;
   
     cupt = snpm[num] ;
     sprintf(cupt -> ID, "fake-%d:%d", xc, nfake) ;
     cupt -> estgenpos = cupt -> genpos = fakedis ; 
     xl = snpm[num-1] -> genpos ;
     al =  snpm[num-1] -> physpos ;
     xr =  realdis ;          ;
     ar =  physpos ;
     cupt -> physpos = interp(xl, xr, fakedis, al, ar) ;
     cupt -> markernum = num ;
     cupt -> isfake = YES ;
     cupt -> chrom  = xc ;
     fakedis += spacing ;
     ++num ;
     ++nfake ;
   }
   ++cnum ;
   cupt = snpm[num] ;
   strcpy(cupt -> ID, sname) ;
   cupt -> estgenpos = cupt -> genpos = realdis ;                          
   cupt -> physpos = physpos ;
   cupt -> markernum = num ;
   cupt -> isfake = NO ;
   cupt -> isrfake = sdpt -> isrfake ; 
   cupt -> chrom  = xc ;
   cupt -> tagnumber = inputrow ; // just used for pedfile 
   if (inputrow >=0) {
    pedcols[inputrow] = num ;
   }
 

   n0 = sdpt->nn[0] ;
   n1 = sdpt->nn[1] ;
   fraw = mknn(nn, n0, n1) ;
   copyiarr(nn, cupt->popA_nn, 2) ;
   cupt -> PopAftrue = cupt -> PopA_freq = fraw ;
   cupt -> adm_PopAftrue = cupt -> adm_PopA_freq = fraw ;

   cupt -> alleles[0] = sdpt -> alleles[0] ;
   cupt -> alleles[1] = sdpt -> alleles[1] ;

   n0 = sdpt->nn[2] ;
   n1 = sdpt->nn[3] ;
   fraw = mknn(nn, n0, n1) ;
   copyiarr(nn, cupt->popB_nn, 2) ;
   cupt -> PopBftrue = cupt -> PopB_freq = fraw ;
   cupt -> adm_PopBftrue = cupt -> adm_PopB_freq = fraw ;
   ++num ;
  }
// now make list of ignored snps used by loadgeno for check
  numsnps = num ;
  nignore = 0 ;
  for (k=0; k< nreal ; k++) {  
   indx = snpindx[k] ;
   sdpt = snpraw[indx] ;
   if (sdpt->ignore == NO) continue ;
   inputrow = sdpt -> inputrow ;
   chrom = sdpt -> chrom ;
   sname = sdpt -> ID ;
   realdis = sdpt -> gpos ;           
   physpos = sdpt -> ppos ;
   cupt = snpm[num] ;
   strncpy(cupt -> ID, sname, IDSIZE-1) ; 
   cupt -> genpos = realdis ;                          
   cupt -> physpos = physpos ;
   cupt -> markernum = num ;
   cupt -> isfake = NO ;
   cupt -> ignore = YES ;
   cupt -> chrom  = chrom ;
   cupt -> tagnumber = inputrow ; // just used for pedfile 
   if (inputrow >=0) {
    pedcols[inputrow] = num ;
   }
   ++nignore ; 
   ++num ;
  }
  *numignore = nignore ;
  return numsnps ;
}

double interp (double l, double r, double x, double al, double ar)
// linearly interp ;
{
  double y, y1, y2 ;
  y = (r-l) ;
  if (y==0.0) return 0.5*(al+ar) ;
  y1 = (r-x)/y ;
  y2 = (x-l)/y ;
  return y1*al +y2*ar ;
}

int getindivs(char *indivfname, Indiv ***indmarkpt) 
// returns number of indivs  
{

 static Indiv **indivmarkers ;
 int nindiv, i ;

 if (indivfname == NULL) fatalx("(getindivs) NULL indivfname\n") ;
 nindiv = getsizex(indivfname) ;
 if (nindiv <= 0) fatalx("no indivs found: indivname: %s\n", indivfname) ;
 ZALLOC(indivmarkers, nindiv, Indiv *) ;

 for (i=0; i<nindiv; i++) {  
  ZALLOC(indivmarkers[i], 1, Indiv) ;
 }
 clearind(indivmarkers, nindiv) ;
 *indmarkpt = indivmarkers ;
 readinddata(indivmarkers, indivfname) ;
 return nindiv ;
}

int readinddata(Indiv **indivmarkers, char *fname) 
{
 char line[MAXSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0, k ;
 int skipit ;
 Indiv *indx ;

 FILE *fff ;

 if (ispedfile(fname)) {  
  plinkinputmode = YES ;
  return  readindpeddata(indivmarkers, fname) ;
 }

 openit(fname, &fff, "r") ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') skipit = YES ;
   if (strcmp(sx,"SNP_ID") == 0) skipit = YES ;
   if (strcmp(sx,"Indiv_ID") == 0) skipit = YES ;
   if (skipit == NO) { 
    if (nsplit<3) fatalx("%s bad line: %s", fname, line) ;
    indx = indivmarkers[num] ;
    if (strlen(sx)>=IDSIZE) fatalx("ID too long: %s\n", sx) ;
    strcpy(indx->ID, sx) ;
    indx -> idnum = num ;
    sx = spt[1] ;
    indx -> gender = sx[0] ;
    indx -> affstatus = indx -> ignore = NO ;
    sx = spt[2] ;
    if (strcmp(sx, "Ignore") == 0) indx->ignore = YES ;
    if ((qtmode) && (!indx->ignore)) { 
     indx -> egroup = strdup("Case") ;
     indx -> qval = indx -> rawqval = atof(sx) ;
    }
    else {
     indx -> egroup = strdup(sx) ;  
    }
// affstatus set by setstatus  
    ++num ;
   }
   freeup(spt, nsplit) ;
   continue ;
 }

 fclose(fff) ;
 return num ;
}

int readindpeddata(Indiv **indivmarkers, char *fname) 
///
{
 char *line ;
 char *spt[MAXFF], *sx, *sx0, gender ;
 int nsplit, num=0, k ;
 int skipit ;
 Indiv *indx ;
 int nindiv ; 
 int maxnsplit = 0 ;
 char nnbuff[IDSIZE] ;

 FILE *fff ;

 maxgenolinelength = maxlinelength(fname) ;
 ZALLOC(line, maxgenolinelength+1, char) ;
 openit(fname, &fff, "r") ;

 while (fgets(line, maxgenolinelength, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx0 = sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') skipit = YES ;
   if (skipit == NO) { 
    if (nsplit<6) fatalx("%s bad line: %s", fname, line) ;
    indx = indivmarkers[num] ;
    if (strlen(sx)>=IDSIZE) fatalx("ID too long: %s\n", sx) ;
    maxnsplit = MAX(maxnsplit, nsplit) ;

    sx = spt[1] ;
    pedname(nnbuff, sx0, sx) ;
    strcpy(indx->ID, nnbuff) ;
    indx -> idnum = num ;
    sx = spt[4] ;
    k = atoi(sx) ;
    gender = 'U' ;
    if (k==1) gender = 'M' ;
    if (k==2) gender = 'F' ;
    indx -> gender = gender ;
    indx -> affstatus = indx -> ignore = NO ;

    sx = spt[5] ;
    if (qtmode) {
     indx -> egroup = strdup("Case") ;
     indx -> qval = indx -> rawqval = atof(sx) ;
    }

    else {
     k = 99 ;
     if (strcmp(sx, "-9") == 0) k = -9 ;
     if (strcmp(sx, "9") == 0) k = 9 ;
     if (strcmp(sx, "0") == 0) k = 0 ;
     if (strcmp(sx, "1") == 0) k = 1 ;
     if (strcmp(sx, "2") == 0) k = 2 ;
     switch (k) { 
      case 9:  
        indx -> ignore = YES ;
        printf("%s ignored\n", indx -> ID) ;
        break ;
      case -9:  
        indx -> ignore = YES ;
        printf("%s ignored\n", indx -> ID) ;
        break ;
      case 0:  
        indx -> ignore = YES ;
        printf("%s ignored\n", indx -> ID) ;
        break ;
      case 1: 
       indx -> egroup = strdup("Control") ;
       break ; 
      case 2: 
       indx -> egroup = strdup("Case") ;
       break ;
      default:  
       indx -> egroup = strdup(sx) ;
     }

    }

// affstatus set by setstatus  
    ++num ;
   }
   freeup(spt, nsplit) ;
   continue ;
 }

 if (maxnsplit<8) maxgenolinelength = -1 ;
 free(line) ;

 fclose(fff) ;
 return num ;
}
void pedname(char *cbuff, char *sx0, char *sx1) 
{
  int l0 , l1, ll ;

  l0 = strlen(sx0) ;
  l1 = strlen(sx1) ;
  ll = l0 +l1 + 1 ; 
  if (familynames == NO) ll = l1 ;
  if (ll>=IDSIZE) {
   fatalx("idnames too long %s %s ll: %d limit: %d\n", sx0, sx1, ll, IDSIZE-1) ;
  }
  if (familynames == YES) {
   strcpy(cbuff, sx0) ; 
   cbuff[l0] = ':' ; 
   strcpy(cbuff+l0+1, sx1) ;
   return ;
  }
  strcpy(cbuff, sx1) ;

}
  
int readtldata(Indiv **indivmarkers, int numindivs, char *inddataname)  
{ 
 char line[MAXSTR] ; 
 char *spt[MAXFF], *sx ; 
 int nsplit, num=0, k, ind ; 
 int skipit ; 
 Indiv *indx ; 
 double y ; 
 double gg[3] ;  
 
 FILE *fff ; 
 openit(inddataname, &fff, "r") ; 
 while (fgets(line, MAXSTR, fff) != NULL) { 
   nsplit = splitup(line, spt, MAXFF) ;  
   if (nsplit == 0) continue ; 
   sx = spt[0] ; 
   skipit = NO ; 
   if (strcmp(sx, "Indiv_Index") == 0) {
// hack.  thetafile should be output with leading ## 
    freeup(spt, nsplit) ; 
    continue ; 
   }
   if (sx[0] == '#') {                
    freeup(spt, nsplit) ; 
    continue ; 
   }
    if (nsplit<8) fatalx("%s bad line: %s", inddataname, line) ; 
    sx = spt[1] ;
    ind = indindex(indivmarkers, numindivs, sx)  ; 
    if (ind<0) fatalx("(readtldata) indiv: %s not found \n", sx) ; 
    indx = indivmarkers[ind] ; 
    indx -> theta_mode = atof(spt[3]) ;
    indx -> lambda_mode  = atof(spt[7]) ;
    freeup(spt, nsplit) ; 
    continue ; 
 } 
  
 fclose(fff) ; 
 return num ; 
} 

int readfreqdata(SNP **snpm, int numsnps, char *inddataname)  
{ 
 char line[MAXSTR] ; 
 char *spt[MAXFF], *sx ; 
 int nsplit, num=0, k, ind ; 
 int skipit ; 
 SNP *cupt ; 
 
 FILE *fff ; 
 openit(inddataname, &fff, "r") ; 
 while (fgets(line, MAXSTR, fff) != NULL) { 
   nsplit = splitup(line, spt, MAXFF) ;  
   if (nsplit == 0) continue ; 
   sx = spt[0] ; 
   skipit = NO ; 
   if (sx[0] == '#') {                
    freeup(spt, nsplit) ; 
    continue ; 
   }
    if (nsplit<6) fatalx("%s bad line: %s", inddataname, line) ; 
    sx = spt[2] ;
    ind = snpindex(snpm, numsnps, sx)  ; 
    if (ind<0) fatalx("(readfreqdata) snp %s not found \n", sx) ; 
    cupt = snpm[ind] ; 
    cupt -> adm_PopA_freq = cupt -> PopA_freq  = atof(spt[3]) ;
    cupt -> adm_PopB_freq = cupt -> PopB_freq  = atof(spt[5]) ;

    freeup(spt, nsplit) ; 
    ++num ;
    continue ; 
 } 
  
 fclose(fff) ; 
 return num ; 
} 



int setstatus(Indiv **indm, int numindivs, char *smatch)   
// return number set 
// smatch = NULL => set everything 
{ 
  return setstatusv(indm, numindivs, smatch, YES) ;
}

int setstatusv(Indiv **indm, int numindivs, char *smatch, int val)   
// return number set 
// smatch = NULL => set everything 
{ 
  int i, n=0 ; 
  Indiv *indx ; 
  char *sx ; 
  for (i=0; i<numindivs; i++) { 
   indx = indm[i] ; 
   if (indx -> ignore) continue ;
   sx = indx -> egroup ;
   if (smatch == NULL) {
    ++n ; 
    indx->affstatus = val ; 
    continue ;
   } 
   if (strcmp(sx, smatch) == 0) {  
    ++n ; 
    indx->affstatus += val ; 
   } 
  } 
  return n ; 
} 


int getgenos(char *genoname, SNP **snpmarkers, Indiv **indivmarkers, 
 int numsnps, int numindivs, int nignore) 
{
// read genofile.  Use hashtable to improve search 
// if genofile is gzipped decompress to trashdir
 char *gname, *genotmp = NULL ;
 ENTRY *hashlist, *iteml  ;
 ENTRY item1 ;
 int k, num, indiv, lgt ;
 int val ;
 char *basept = 0 ;
 int bigoff ;

// we use a trick: want to store k 
// store basept + k instead 
// basept + k + bigoff for individual ID 

 SNP *cupt ;
 Indiv *indx ;

 char line[MAXSTR], cmd[MAXSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, nsnp ;
 int skipit, kret, tpackmode, teigenstratmode ;

 FILE *fff ;
 int gnlen, ngenos=0 ;

 double y ;
 char *pbuff ;

 if (genoname == NULL) fatalx("(getgenos) NULL genoname\n") ; 
 gname = genoname ;
 gnlen = strlen(genoname) ;
 if (strcmp(genoname+gnlen-3, ".gz") == 0)  {  
  makedir(trashdir) ; 
  sprintf(line, "%s/genotmp:%d", trashdir, getpid()) ;
  genotmp = strdup(line) ;
  sprintf(cmd, "gunzip -c %s > %s", genoname, genotmp) ;
  printf("unzip cmd: %s\n", cmd) ;
  system (cmd) ;
  kret = system (cmd) ;
  if (kret<0) {  
   perror("gunzip failed\n") ;
   fatalx("gunzip failed... probably out of disk space\n") ;
  }
  printf("geno file unzipped\n") ;
  gname = genotmp ;
 }

 if (ispedfile(gname))   {
   if (pedcols == NULL) fatalx("pedcols not allocated (no map file ?)") ;
   getpedgenos(genoname, snpmarkers, indivmarkers, 
     numsnps,  numindivs,  nignore)  ;
   freeped()  ;
   return ; 
 }  

 if (isbedfile(gname))   {
   return  getbedgenos(genoname, snpmarkers, indivmarkers, 
     numsnps,  numindivs,  nignore)  ;
 }  

 tpackmode = ispack(gname) ;
 

 nsnp = numsnps + nignore ;
 if (tpackmode) { 
   inpack(gname, snpmarkers, indivmarkers,  nsnp, numindivs)  ; 
   for (k=0; k<nsnp; k++)  { 
    cupt = snpmarkers[k] ;
    if (cupt -> ignore) continue ;
    if ((cupt -> isfake) && (!(cupt -> isrfake))) continue ;
    cupt -> ngtypes = numindivs ; 
    if (cupt -> gtypes == NULL) ZALLOC(cupt -> gtypes, 1, int) ; 
   }
   packmode = YES ;
   return nsnp*numindivs ;
 }
 teigenstratmode = iseigenstrat(gname) ;
 if (teigenstratmode) { packmode = YES ;
  ineigenstrat(gname, snpmarkers, indivmarkers,  nsnp, numindivs)  ; 
  freeped() ;
  return nsnp*numindivs ;
 }

 

 y = (double) (numindivs * 2) / (8 * (double) sizeof (char)) ;  
 rlen = nnint(ceil(y)) ;
 packlen = rlen*numsnps ;
 if (packmode) {
   ZALLOC(packgenos, packlen, char) ;
   pbuff = packgenos ;
   clearepath(packgenos) ;
 }

 num = nsnp + numindivs ;
 xhcreate(5*num) ;
 ZALLOC(hashlist, num, ENTRY) ;
 bigoff = nsnp + 100 ;

 for (k=0; k<nsnp; k++)  { 
  cupt = snpmarkers[k] ;
  if ((cupt -> isfake) && (!(cupt -> isrfake))) continue ;
  iteml = hashlist+k ;
  iteml -> key = cupt->ID  ;
  iteml -> data = basept +k ;  
  if (xhsearch(*iteml, FIND)  != NULL) 
   fatalx("duplicate ID: %s\n", iteml -> key) ;
  (void) xhsearch(*iteml, ENTER) ;
 }

 for (k=0; k<numindivs; k++)  { 
  indx = indivmarkers[k] ;
  iteml = hashlist+numsnps+k ;
  iteml -> key = indx->ID  ;
  iteml -> data = basept+k+bigoff ;  
  if (xhsearch(*iteml, FIND)  != NULL) 
   fatalx("duplicate ID: %s\n", iteml -> key) ;
  (void) xhsearch(*iteml, ENTER) ;
 }
 openit(gname, &fff, "r") ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') skipit = YES ;
   if (strcmp(sx,"SNP_ID") == 0) skipit = YES ;
   if (strcmp(sx,"Indiv_ID") == 0) skipit = YES ;
   if (skipit == NO) { 
    item1.key = spt[0] ;
    iteml = xhsearch(item1, FIND) ; 
    if (iteml == NULL) {
     fatalx("(genotypes) bad ID (SNP): %s\n", line) ;
    }
    k = (int) iteml->data - (int) basept ;
    cupt = snpmarkers[k] ;
    if (cupt -> ignore) {
      freeup(spt, nsplit) ;
      continue ;
    }
    item1.key = spt[1] ;
    iteml = xhsearch(item1, FIND) ; 
    if (iteml == NULL) {
     fatalx("(genotypes) bad ID: (Indiv) %s\n", line) ;
    }
    indiv = (int) iteml->data - (int) basept ;
    indiv -= bigoff ;
    val = atoi(spt[2]) ;

    indx = indivmarkers[indiv] ;
    if (indx->ignore)  val = -1 ;
    if (checkxval(cupt, indx, val) == NO) val = -1 ;
    if (val>2) {  
     printf("*** warning invalid genotype: %s %s %d\n", 
       cupt -> ID, indx -> ID, val) ;  
       val = -1 ;
    }

    if (cupt -> ngtypes == 0) { 
     if (packmode == NO) {
      ZALLOC(cupt -> gtypes, numindivs, int) ;
     }
     else  {
      ZALLOC(cupt -> gtypes, 1, int) ;
      cupt -> pbuff = pbuff ;  
      pbuff += rlen ;
     }
     cupt -> ngtypes = numindivs ;
     for (k=0; k<numindivs; ++k) {
      putgtypes(cupt, k, -1) ;
     }
    }
    putgtypes(cupt, indiv, val) ;
    ++ngenos ;
   }
   freeup(spt, nsplit) ;
 }
 fclose(fff) ;
 free(hashlist) ;
 xhdestroy() ;
 if (genotmp != NULL) { 
   unlink(gname) ;
 }
 /* printf("genotype file processed\n") ; */
 freeped() ; 
 return ngenos ;
}

void freeped() 
// destructor for pedcols 
{
 if (pedcols == NULL) return ; 
 free(pedcols) ;
 pedcols = NULL ;
}

int checkxval(SNP *cupt, Indiv *indx, int val) 
// check Male X marker not het
{
   if (cupt -> chrom != 23) return YES ;
   if (indx -> gender != 'M') return YES ;
   if (val != 1) return YES ;  
   if (malexhet)  return YES ;
   return NO; 
}

void clearsnp(SNP *cupt) 
{

   cupt -> PopA_freq =
   cupt -> PopB_freq = -1 ;
   cupt -> adm_PopA_freq = 
   cupt -> adm_PopB_freq = -1 ; 
   cupt -> estgenpos = 0 ;
   cupt -> genpos = 0 ;
   cupt -> physpos = 0 ;
   cupt -> ngtypes = 0 ;
   cupt -> pbuff = NULL ;
   cupt -> ebuff = NULL ;
   cupt -> gtypes = NULL ;
   cupt -> modelscores = NULL ;
   cupt -> totmodelscores = NULL ;
   cupt -> isfake = NO ; 
   cupt -> ignore = NO ; 
   cupt -> isrfake = NO ; 
   cupt -> estdis = 0 ; 
   cupt -> dis = 0 ; 
   cupt -> esum = 0 ; 
   cupt -> lsum = 0 ;
   cupt -> gpsum = 0 ; 
   cupt -> gpnum = 0 ;
   cupt -> pcupt = NULL ;
   cupt -> tagnumber = -1 ;
   cclear((unsigned char *) cupt -> alleles, CNULL, 2)  ;
}

void rmsnps(SNP **snpm, int numsnps, int nignore) 
{
   int i ;
   SNP *cupt ;


/**
   for (i=0; i<numsnps; i++)  {  
    cupt = snpm[i] ;
    if (cupt -> gtypes == NULL) cupt -> ignore = YES ;
    if (cupt -> ngtypes == 0) cupt -> ignore = YES ;
   }
*/


   for (i=0; i< nignore; i++) {  
    cupt = snpm[numsnps+i] ;
    if (cupt -> modelscores != NULL) { 
     free(cupt->modelscores) ;
    }
    if (cupt -> totmodelscores != NULL) { 
     free(cupt->totmodelscores) ;
    }
    free(cupt) ;
   }
}

void
clearind(Indiv **indm, int numind)   
{
   Indiv *indx ;  
   double theta ; 
   int i ;

   for (i=0; i<numind; i++)  {
    indx = indm[i] ;
    indx -> egroup = NULL ;
    indx -> affstatus = indx -> ignore = NO ;
    indx -> gender = 'U' ;
    indx = indm[i] ;
    indx -> Xtheta_mode = indx->theta_mode = a1/(a1+b1) ;
    indx -> Xlambda_mode = indx -> lambda_mode = lp1/lp2 ;
    indx -> thetatrue = -1.0 ;  // silly value
    indx -> qval = indx -> rawqval = 0.0 ;
   }
   cleartg(indm, numind) ;
}

void cleartg(Indiv **indm, int nind) 
{
    int i ;
    Indiv *indx ;

      for (i=0; i< nind; i++) {  
       indx = indm[i] ;
       vzero(indx -> totgamms, 3) ;
       indx -> totscore = 0.0 ; 
      }

}


double mknn(int *nn, int  n0, int n1) 
{

  double x ;
  int t ;

  nn[0] = n0 + 1 ;       
  nn[1] = n1 + 1 ;       

// no clipping.  (Old code clipped here)
  t = intsum(nn,2) ;
  x = ((double) nn[0]) / (double) t ;

  return x ;
}

void
setug(Indiv **indm, int numind, char gender)   
{

   Indiv *indx ;  
   double theta ; 
   int i ;

   for (i=0; i<numind; i++)  {
    indx = indm[i] ;
    if (indx -> gender == 'U') indx -> gender = gender ;
   }

}


void dobadsnps(SNPDATA **snpraw, int nreal, char *badsnpname) 
{

  FILE *fff ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *ss ;
  int indx, nsplit, n ;

  if (badsnpname == NULL) return ;
  openit (badsnpname, &fff, "r") ;

  while (fgets(line, MAXSTR, fff) != NULL)  {
    nsplit = splitup(line, spt, MAXFF) ; 
    if (nsplit==0) continue ;
    if (spt[0][0] == '#') { 
     freeup(spt, nsplit) ;
     continue ;
    }
    indx = snprawindex(snpraw, nreal, spt[0]) ;
    if (indx >=0) {
     snpraw[indx] -> ignore = YES ;
     if ((nsplit >=2) && (checkfake(spt[1]))) {  
       snpraw[indx] -> ignore = NO ;
       snpraw[indx] -> isrfake = YES ;
     }
    }
    freeup(spt, nsplit) ;
  }
  fclose (fff) ;
}
int checkfake(char *ss) 
// yes if string ss is "Fake" 
// ss is overwritten
{
  ss[0] = tolower(ss[0]) ;
  if (strcmp(ss,"fake") == 0) return YES ;
  return NO ;

}
int snprawindex(SNPDATA **snpraw, int nreal, char *sname) 
{
    int i ;

    for (i=0; i<nreal; ++i)  {  
      if (strcmp(sname, snpraw[i] -> ID) == 0) return i ;
    }

    return -1 ;
}

void printsnps(char *snpoutfilename, SNP **snpm, int num, 
  Indiv **indm, int printfake, int printvalids) 
{ 
  int i, ppos ;
  SNP *cupt  ;
  char ss[MAXSTR] ;
  FILE *xfile ;
  int numvcase, numvcontrol ;
  char c ;

  if ((snpoutfilename != NULL) && (strcmp(snpoutfilename, "NULL") == 0))  return ;
  if (snpoutfilename != NULL)  {
   openit(snpoutfilename, &xfile, "w") ;
  }
  else xfile = stdout ;

  if( tersemode == NO) {
    fprintf(xfile,"\n");
    fprintf(xfile,"###DETAILS ABOUT THE MARKERS\n");
    fprintf(xfile,"##Gen_Pos: genetic position, Phys_pos: Physical position\n");
    fprintf(xfile,"##PopA_vart: Parental PopA variant allele count, PopA_ref: Parental PopA reference allele count\n");
    fprintf(xfile,"##PopB_vart: Parental PopB variant allele count, PopB_ref:Parental PopB reference allele count\n");   

    fprintf(xfile, "\n") ;
    fprintf(xfile,"%15s %5s %10s %15s","#SNP_Id","Chr_Num","Gen_Pos","Phys_Pos") ;
    fprintf(xfile,  "%11s %9s %9s %8s" ,"PopA_vart","PopA_ref","PopB_vart","PopB_ref");
    fprintf(xfile,  "%7s %8s" ,"Case_Cnt","Ctl_Cnt");
    fprintf(xfile, "\n") ;
  }
  for (i = 0; i <num; ++i) {  
     cupt = snpm[i] ;
     if (!printfake && (ignoresnp(cupt))) continue ;
     if (!printfake && (cupt -> isrfake)) continue ;
     fprintf(xfile, "%15s %3d ", cupt->ID, cupt -> chrom) ;
     fprintf(xfile, "%15.6f %15.0f",  cupt -> genpos, cupt -> physpos) ;

     if (tersemode) { 
      printalleles(cupt, xfile) ;
      fprintf(xfile, "\n") ;
      continue ;
     }

     fprintf(xfile, " %8d ", cupt -> popA_nn[0]) ;
     fprintf(xfile, "%8d ", cupt -> popA_nn[1]) ;
     fprintf(xfile, "%8d ", cupt -> popB_nn[0]) ;
     fprintf(xfile, "%8d", cupt -> popB_nn[1]) ;
     if (!printvalids) { 
      printalleles(cupt, xfile) ;
      fprintf(xfile, "\n") ;
      continue ;
     }
     numvcase = numvalidgtx(indm, cupt, 1) ;
     numvcontrol = numvalidgtx(indm, cupt, 0)  ;
     fprintf(xfile, "   %6d %6d",numvcase, numvcontrol) ;
     printalleles(cupt, xfile) ;
     fprintf(xfile, "\n") ;
  }
  if (snpoutfilename != NULL) 
    fclose(xfile) ;
}
void printalleles(SNP *cupt, FILE *fff)  
{

  char c ;
  if ((c = cupt -> alleles[0]) != CNULL) fprintf(fff, " %c", c) ;
  if ((c = cupt -> alleles[1]) != CNULL) fprintf(fff, " %c", c) ;

}

void printdata(char *genooutfilename, char *indoutfilename,
  SNP **snpm, Indiv **indiv, int numsnps, int numind, int packem)
{

  FILE *gfile, *ifile;
  int i,j, t;
  SNP *cupt;
  Indiv *indx;
  char ss[MAXSTR];
  char *gfilename ;
  int dogenos = YES ;

  if (packem) 
   printf("packedancestrymap output\n") ;
  else 
   printf("ancestrymap output\n") ;

  if ((genooutfilename != NULL) && (strcmp(genooutfilename, "NULL") == 0))  dogenos = NO ;
  if (genooutfilename == NULL) dogenos = NO ;

  if (dogenos) {
   gfilename = genooutfilename ;
   if (packem) { 
    outpack(genooutfilename, snpm, indiv, numsnps, numind) ;
     gfilename = NULL ;
    }

   if (gfilename != NULL)  {
    openit(gfilename, &gfile, "w") ;
    if(tersemode == NO) fprintf(gfile,"#SNP_ID,INDIV_ID,VART_ALLELE_CNT\n");
   }

   for (i = 0; i< numsnps; i++)  {
      if (gfilename == NULL) break ;
      cupt= snpm[i];
      if (cupt -> ignore) continue ;
      if (cupt -> isfake) continue ;
      if (cupt -> ngtypes == 0)  continue ;
      for(j=0; j < cupt->ngtypes; j++)
	{
	  indx = indiv[j];
          if (indx -> ignore) continue  ;  
	  fprintf(gfile,"%20s %20s %3d\n",cupt->ID,indx->ID, getgtypes(cupt,j)) ;
	}
     }
   

   if(gfilename != NULL)
    fclose(gfile);

  }
  
  if(indoutfilename == NULL)
    return;
  if ((indoutfilename != NULL) && (strcmp(indoutfilename, "NULL") == 0))  return ;
  if (indoutfilename != NULL) 
    openit(indoutfilename, &ifile, "w") ;

  /* fprintf(ifile,"#INDIV,GENDER,POPULATION\n"); */
  for(i = 0; i< numind; i++)
    {
      indx = indiv[i];
      if (indx->ignore) continue ;            
      strcpy(ss, indx -> egroup) ;
      if ((qtmode) && (!indx->ignore)) {  
        sprintf(ss, "%9.3f", indx -> rawqval) ;
      }
      if (tersemode) { 
       fprintf(ifile,"%20s %c %10s",indx->ID, indx->gender,ss);
       fprintf(ifile,"\n") ;
       continue ;
      }
      t = numvalids(indx, snpm, 0, numsnps-1) ;
      fprintf(ifile,"%20s %c %10s %5d\n",indx->ID, indx->gender,ss, t);
    }

  if(indoutfilename != NULL)
    fclose(ifile);

}

int readgdata(Indiv **indivmarkers, int numindivs, char *gname)  
     // only needed for logreg 
     // not correct for X chromosome 
     // Needs correction  for males 
{ 
 char line[MAXSTR] ; 
 char *spt[MAXFF], *sx ; 
 int nsplit, num=0, k, ind ; 
 int skipit ; 
 Indiv *indx ; 
 double y ; 
 double gg[3] ;  
 
 FILE *fff ; 
 openit(gname, &fff, "r") ; 
 while (fgets(line, MAXSTR, fff) != NULL) { 
   nsplit = splitup(line, spt, MAXFF) ;  
   if (nsplit == 0) continue ; 
   sx = spt[0] ; 
   skipit = NO ; 
   if (sx[0] == '#') skipit = YES ; 
   if (strcmp(sx,"SNP_ID") == 0) skipit = YES ; 
   if (strcmp(sx,"Indiv_ID") == 0) skipit = YES ; 
   if (skipit == NO) {  
    if (nsplit<4) fatalx("%s bad line: %s", gname, line) ; 
    ind = indindex(indivmarkers, numindivs, sx)  ; 
    if (ind<0) fatalx("(readgdata) indiv: %s not found \n", sx) ; 
    indx = indivmarkers[ind] ; 
    for (k=0; k<3; k++) {  
     gg[k] = atof(spt[k+1]) ; 
    } 
    y = asum(gg, 3) ;  
    vst(gg, gg, 1.0/y, 3) ; 
    y = 0.5*(gg[1]+gg[2]) ;  /* est PopB ancestry */ 
    indx -> thetatrue = y ;   
    copyarr(gg, indx -> totgamms, 3) ; 
   } 
   freeup(spt, nsplit) ; 
   continue ; 
 } 
  
 fclose(fff) ; 
 return num ; 
} 
int numvalidind(Indiv **indivmarkers, int  numind)  
{ 
  int i ;              
  int nvalids = 0 ;
  Indiv *indx ;

  for (i=0; i< numind;  i++) {  
   indx = indivmarkers[i] ; 
   if (indx -> ignore) continue ;
   ++nvalids ;
  }
  return nvalids ; 
} 
int numvalidgt(Indiv **indivmarkers, SNP *cupt)  
     // like numvalidgtypes but tests ignore   
{ 
  int n, k, nvalids ; 
  if (cupt -> gtypes == NULL) return 0 ; 
  nvalids = 0 ; 
  n = cupt -> ngtypes ; 
  for (k=0; k<n; k++) {   
    if (indivmarkers[k] -> ignore)  continue ; 
    if (getgtypes(cupt, k) >=0) ++nvalids ; 
  } 
  return nvalids ; 
} 
int numvalidgtx(Indiv **indivmarkers, SNP *cupt, int affst)  
// like numvalidgtypes but tests ignore   counts only when status=affst
{ 
  int n, k, nvalids ; 
  if (cupt -> gtypes == NULL) return 0 ; 
  nvalids = 0 ; 
  n = cupt -> ngtypes ; 
  for (k=0; k<n; k++) {   
    if (indivmarkers[k] -> ignore)  continue ; 
    if (indivmarkers[k] -> affstatus != affst) continue ;
    if (getgtypes(cupt, k) >=0) ++nvalids ; 
  } 
  return nvalids ; 
} 
int getweights(char *fname, SNP **snpm, int numsnps)  
// number of real lines 
{
 char line[MAXSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0 ;
 int skipit, k ;
 double weight ;

 FILE *fff ;
 for (k=0; k<numsnps; ++k) { 
  snpm[k] -> weight = 1.0 ;
 }
 openit(fname, &fff, "r") ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) { 
     continue ;
   }
   sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') skipit = YES  ;
   if (strcmp(sx,"SNP_ID") == 0) skipit = YES ;
   if (strcmp(sx,"Indiv_ID") == 0) skipit = YES ;
   k = snpindex(snpm, numsnps, sx) ;
   if (k<0) skipit = YES ;
   if (skipit == NO) {
    if (nsplit >1) {  
     sx = spt[0] ;
     weight =  atof(sx) ;
     snpm[k] -> weight = weight ;
     printf("weight set: %20s %9.3f\n", snpm[k] -> ID, weight) ;
     ++num ;
    }
   }
   freeup(spt, nsplit) ;
   continue ;
 }
 fclose(fff) ;
 fflush(stdout) ;
 return num ;
}
void
outpack(char *genooutfilename, SNP **snpm, Indiv **indiv, int numsnps, int numind) 
{
   char **arrx ;  
   int n, num, ihash, shash, i, g, j, k ;
   int nind , nsnp, irec ;
   Indiv *indx ; 
   SNP *cupt ;
   double y ;
   unsigned char *buff  ;
   int fdes, ret ;
   char *packit ;

   n = numind ;
   ZALLOC(arrx, n, char *) ;

   num = 0 ;
   for (i=0; i<n ; i++)  {  
    indx = indiv[i] ;  
    if (indx -> ignore) continue ;
    arrx[num] = strdup(indx -> ID) ;
    ++num ;
   }
   ihash = hasharr(arrx, num) ;
   nind= num ;


   freeup(arrx, num) ;  
   free(arrx) ;
   
   n = numsnps ;
   ZALLOC(arrx, n, char *) ;
   num = 0 ;
   for (i=0; i<n ; i++)  {  
    cupt = snpm[i] ;  
    if (cupt -> ignore) continue ;
    if (cupt -> ngtypes == 0) continue ;
    arrx[num] = strdup(cupt -> ID) ;
    ++num ;
   }
   shash = hasharr(arrx, num) ;
   nsnp = num ;
   freeup(arrx, num) ;  
   free(arrx) ;

   // printf("ihash:  %x   shash: %x\n", ihash, shash) ; 
   y = (double) (nind * 2) / (8 * (double) sizeof (char)) ;  
   rlen = nnint(ceil(y)) ;
   rlen = MAX(rlen, 48)  ;  
   // printf("nind: %d  rlen:  %d\n", nind, rlen) ;
   ZALLOC(buff, rlen, unsigned char) ;
   sprintf((char *) buff,"GENO %6d %6d %x %x", nind, nsnp, ihash, shash) ;

   ridfile(genooutfilename) ;
   fdes = open(genooutfilename, O_CREAT | O_TRUNC | O_RDWR, 0666);
 
   if (fdes<0) {
    perror("bad genoout") ;
    fatalx("open failed for %s\n", genooutfilename) ;
   }
   if (verbose) 
    printf("file %s opened\n", genooutfilename) ;

    ret = write(fdes, buff, rlen) ;
    if (ret<0) {  
     perror("write failure") ;
     fatalx("(outpack) bad write") ;
    } 
    ZALLOC(packit, nsnp*rlen, char) ;

    irec = 1; 
    for (i=0; i<numsnps ; i++)  {  
     cupt = snpm[i] ;  
     if (cupt -> ignore) continue ;
     if (cupt -> ngtypes == 0) continue ;
     cclear((unsigned char *) buff, 0X00, rlen) ;  
     num = 0 ;
     for (j=0; j< numind; j++) { 
      indx = indiv[j] ;  
      if (indx -> ignore) continue ;
      g = getgtypes(cupt, j) ;
      if (g<0) g=3 ;
      wbuff( buff, num, g) ;
      ++num ;
     }
     ret = write(fdes, buff, rlen) ;
     if (ret<0) {  
      perror("write failure") ;
      fatalx("(outpack) bad write") ;
     } 
     if (verbose) {
      printf("record: %4d ", irec) ; 
      for (k=0; k<rlen; ++k) {  
       printf(" %02x", (unsigned char) buff[k]) ;
      }
      printf("\n") ;
     }
     ++irec ;
    }
    close(fdes) ;
    free(buff) ;
    // printf("check: %s %d\n", genooutfilename, ispack(genooutfilename)) ;
}
int ispack(char *gname) 
// checks if file is packed gfile 
{
  int fdes, t, ret ;  
  char buff[8] ;

  fdes = open(gname, O_RDONLY) ;
  if (fdes<0) {  
   perror("open failure") ;
   fatalx("(ispack) bad open %s\n", gname) ;
  } 
  t = read(fdes, buff, 8 ) ;
  if (t<0) {  
   perror("read failure") ;
   fatalx("(ispack) bad read") ;
  } 
  close(fdes) ;
  buff[4] = '\0' ;  
  ret = strcmp(buff, "GENO") ;
  if (ret == 0) return YES ;
  return NO ;
  
}
int iseigenstrat(char *gname)  

{

   FILE *fff ;
   char line[MAXSTR] ;
   char *spt[MAXFF], *sx ;
   int nsplit, num=0, k ;


   openit(gname, &fff, "r") ;

   while (fgets(line, MAXSTR, fff) != NULL)  {
    nsplit = splitup(line, spt, MAXFF) ; 
    if (nsplit ==0 ) continue ;
    sx = spt[0] ;
    if (sx[0] == '#')  { 
     freeup(spt, nsplit) ;
     continue ;
    }
    freeup(spt, nsplit) ;
    fclose(fff) ;
    if (nsplit>1) return NO ; 
    return YES ;
   }
   fatalx("(iseigenstrat) no genotyped data found\n") ;

}
void
ineigenstrat(char *gname, SNP **snpm, Indiv **indiv, int numsnps, int numind) 
{
   FILE *fff ;
   char *line, c ;
   char *spt[2], *sx ;
   int nsplit, rownum=0, k, num ;
   int maxstr, maxff = 2 ; 
   int nind, nsnp, len ;
   double y  ;
   unsigned char *buff  ;
   char *packit, *pbuff ;
   int *gtypes, g ;
   SNP *cupt ; 
   Indiv *indx ;


   packmode = YES ;
   maxstr = numind+10 ;
   ZALLOC(line, maxstr, char) ;

   nind = numind ; 
   nsnp = numsnps ;
   ZALLOC(gtypes, nind, int) ;

   y = (double) (nind * 2) / (8 * (double) sizeof (char)) ;  
   rlen = nnint(ceil(y)) ;
   rlen = MAX(rlen, 48)  ;  
   ZALLOC(buff, rlen, unsigned char) ;
   packlen = rlen*nsnp ;
   ZALLOC(packgenos, packlen, char) ;
   clearepath(packgenos) ;

   openit(gname, &fff, "r") ;

   rownum = 0 ;
   pbuff = packgenos  ; 
   while (fgets(line, maxstr, fff) != NULL)  {
    nsplit = splitup(line, spt, maxff) ; 
    if (nsplit ==0 ) continue ;
    sx = spt[0] ;
    if (sx[0] == '#')  { 
     freeup(spt, nsplit) ;
     continue ;
    }
    if (nsplit>1) fatalx("(ineigenstrat) more than 1 field\n") ;
    sx = spt[0] ;
    len = strlen(sx) ;  
    if (len != nind) {
       printf("(ineigenstrat) bad line %d ::%s\n", rownum, line) ;
       fatalx("(ineigenstrat) mismatch line length %d %d\n", len, nind) ;
    }
    if (rownum>=numsnps) fatalx("(ineigenstrat) too many lines in file %d %d\n", rownum, numsnps) ;
    num = pedcols[rownum] ;
    cupt = snpm[num] ; 
    ++rownum ;
    for (k=0; k<len; k++) { 
     sscanf(sx+k, "%c", &c) ;  
     g = -2 ; 
     if (c=='0') g = 0 ;
     if (c=='1') g = 1 ;
     if (c=='2') g = 2 ;
     if (c=='9') g = -1 ;

     if (g==-2) fatalx("(ineigenstrat) bad character %c\n", c) ;

     if (indiv[k] -> ignore) g = -1 ;
     if (checkxval(cupt, indiv[k], g) == NO) g = -1 ;

     if (cupt -> ngtypes == 0) { 
      if (packmode == NO) {
       ZALLOC(cupt -> gtypes, numind, int) ;
      }
      else  {
       ZALLOC(cupt -> gtypes, 1, int) ;
       cupt -> pbuff = pbuff ;  
       pbuff += rlen ;
      }
      cupt -> ngtypes = numind ;
     }
    indx = indiv[k] ;
    if (checkxval(cupt, indx, g) == NO) g = -1 ;
     putgtypes(cupt, k, g) ;
    }
    freeup(spt, nsplit) ;
   }
   if (rownum != numsnps) fatalx("(ineigenstrat) mismatch in numsnps %d and numlines %d\n", numsnps, rownum) ;
   fclose(fff) ;
}

void
inpack(char *gname, SNP **snpm, Indiv **indiv, int numsnps, int numind) 
{
   char **arrx, junk[10] ;  
   int n, num, ihash, shash, i, g, j, k, t ;
   int xihash, xshash, xnsnp, xnind ;
   int nind , nsnp, irec ;
   Indiv *indx ; 
   SNP *cupt ;
   double y ;
   unsigned char *buff  ;
   int fdes, ret ;
   char *packit, *pbuff ;

   n = numind ;
   ZALLOC(arrx, n, char *) ;

   num = 0 ;
   for (i=0; i<n ; i++)  {  
    indx = indiv[i] ;  
    arrx[num] = strdup(indx -> ID) ;
    ++num ;
   }
   ihash = hasharr(arrx, num) ;
   nind= num ;


   freeup(arrx, num) ;  
   free(arrx) ;
   
   n = numsnps ;
   ZALLOC(arrx, n, char *) ;
   num = 0 ;
   for (i=0; i<n ; i++)  {  
    cupt = snpm[i] ;  
    if (cupt -> isfake) continue ;
    arrx[num] = strdup(cupt -> ID) ;
    cupt -> ngtypes = nind ;
    ++num ;
   }
   shash = hasharr(arrx, num) ;
   nsnp = num ;
   freeup(arrx, num) ;  
   free(arrx) ;

   y = (double) (nind * 2) / (8 * (double) sizeof (char)) ;  
   rlen = nnint(ceil(y)) ;
   rlen = MAX(rlen, 48)  ;  
   ZALLOC(buff, rlen, unsigned char) ;

  fdes = open(gname, O_RDONLY) ;
  if (fdes<0) {  
   perror("open failure") ;
   fatalx("(ispack) bad open %s\n", gname) ;
  } 
  t = read(fdes, buff, rlen ) ;
  if (t<0) {  
   perror("read failure") ;
   fatalx("(inpack) bad read") ;
  } 
  sscanf((char *) buff,"GENO %6d %6d %x %x", &xnind, &xnsnp, &xihash, &xshash) ;

  if (xnind != nind) fatalx("OOPS number of individuals %d != %d in input files\n", nind, xnind) ;
  if (xnsnp != nsnp) fatalx("OOPS number of SNPs %d != %d in input files\n", nsnp, xnsnp) ;
  if (xihash != ihash) fatalx("OOPS indiv or snp file has changed since genotype file was created\n") ;
  if (xshash != shash) fatalx("OOPS indiv or snp file has changed since genotype file was created\n") ;

  packlen = rlen*nsnp ;
  ZALLOC(packgenos, packlen, char) ;
  clearepath(packgenos) ;

/**
   printf("packgenos: %x end: %x  len:  %d\n", 
    packgenos, packgenos+packlen-1, packlen) ;
*/

  t = read(fdes, packgenos, packlen ) ;
  if (t<0) {  
   perror("read failure") ;
   fatalx("(inpack) bad data read") ;
  } 
  if (t != packlen) {  
   perror("read failure (length mismatch)") ;
   fatalx("(inpack) bad data read (length mismatch)") ;
  } 
/* now set up pointers into packed data */
   pbuff = packgenos ;
   for (i=0; i<n ; i++)  {  
    cupt = snpm[i] ;  
    if (cupt -> isfake) continue ;
    cupt -> pbuff = pbuff ;
    pbuff += rlen ;
// now check xhets
    for (k=0; k<numind; ++k) {  
     indx = indiv[k] ;  
     g = getgtypes(cupt, k) ;
     if (checkxval(cupt, indx, g) == NO) {
       putgtypes(cupt, k, -1) ;
     }
    }
   }

  free(buff) ;  
  close(fdes) ;
}

void
getsnpsc(char *snpscname, SNP **snpm, int numsnps)
{

   FILE *fff ;
   int  score  ;
   SNP *cupt ;
   char line[MAXSTR] ;
   char *spt[MAXFF], *sx ;
   int nsplit, num=0, k ;
   double y ;


   if (snpscname == NULL)  fatalx("no snpsc file\n") ;
   else openit(snpscname, &fff, "r") ;

   while (fgets(line, MAXSTR, fff) != NULL)  {
    nsplit = splitup(line, spt, MAXFF) ; 
    if (nsplit ==0 ) continue ;
    sx = spt[0] ;
    if (sx[0] == '#')  { 
     freeup(spt, nsplit) ;
     continue ;
    }
    k = snpindex(snpm, numsnps, sx) ;
    if (k<0) {  
     printf("*** warning.  snp %s in snpscname but not in main snp file\n", spt[0]) ;
     freeup(spt, nsplit) ;
     continue ;
    }
    y = atof(spt[1]) ;
    y += .1 * gauss() ; // dither 
    cupt = snpm[k] ; 
    cupt -> score = y ;
    freeup(spt, nsplit) ;
   }

   if (snpscname != NULL)  fclose(fff)  ;  
}

void setepath(SNP **snpm, int nsnps) 
{
 int i ; 
 SNP *cupt ;
 char *pbuff ;

 if (packlen==0) fatalx("(setepath) packlen unset\n") ;
 ZALLOC(packepath, packlen, char) ;
 pbuff = packepath ;
 for (i=0 ; i<nsnps; i++)  {  
   cupt = snpm[i] ;  
   if (cupt -> isfake) continue ;
   cupt -> ebuff = pbuff ;
   pbuff += rlen ;
 }
 clearepath(packepath) ;
}
void clearepath(char *packp) 
{
 cclear((unsigned char *) packp, 0XFF, packlen) ;
}

int getpedgenos(char *gname, SNP **snpmarkers, Indiv **indivmarkers, 
 int numsnps, int numindivs, int nignore) 
{

 int val ;
 int ngenos = 0 ;

 SNP *cupt ;
 Indiv *indx ;

 char *line ;
 char **spt, *sx ;
 char  c ;
 int nsplit, num=0 ;
 int skipit ;
 int numf, snpnumber, nsnp ;
 int k, n, t, i ;
 FILE *fff ;
 int **gcounts, *gvar, *gref ;
 int xvar, xref ;
 int parity, colbase, ncols ; 
 int snpnum ;
 int markernum = -99 ;

/**
 markernum = snpindex(snpmarkers, numsnps, "rs3002685") ;
 if (markernum <0) fatalx("qq1") ;
*/

 if (maxgenolinelength == -1) 
   maxgenolinelength = maxlinelength(gname) ;

// printf("maxlinelen %d\n", maxlinelength(gname)) ;
 ZALLOC(line, maxgenolinelength+1, char) ;

 cleargdata(snpmarkers, numsnps, numindivs) ;
 nsnp = numsnps + nignore ;

 ZALLOC(gcounts, nsnp, int *) ;
 for (i=0; i<nsnp; i++)  { 
   ZALLOC(gcounts[i], 5, int) ;
 }
 genopedcnt(gname, gcounts, nsnp) ;

 ZALLOC(gvar, nsnp, int) ;
 ZALLOC(gref, nsnp, int) ;
 setgref(gcounts, nsnp, gvar, gref) ;

 for (i=0; i<nsnp; ++i) { 
  cupt = snpmarkers[i] ;  
  if (cupt -> alleles[0] != CNULL) { 
   c =  cupt -> alleles[0] ;  
   gvar[i] = xpedval(c) ;
   c =  cupt -> alleles[1] ;  
   gref[i] = xpedval(c) ;
  }
  else {  
   c = x2base(gvar[i]) ;
   cupt -> alleles[0]  = c ;  
   c = x2base(gref[i]) ;
   cupt -> alleles[1]  = c ;  
  }
 }

 numf = 2*nsnp+10 ;
 ZALLOC(spt, numf, char *) ;

 openit(gname, &fff, "r") ;
 while (fgets(line, maxgenolinelength, fff) != NULL)  {
   nsplit = splitup(line, spt, numf) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   skipit = NO ;
   if (sx[0] == '#') {  	
    freeup(spt, nsplit) ;
    continue ;
   }
   if (num==0) {  
    parity = nsplit % 2 ;  
    ncols = nsplit ; 
    colbase = 6 + parity ;
   }
   if (nsplit != ncols) 
    fatalx("bad number of fields %d %d\n", ncols, nsplit) ;

   for (k=colbase ; k < nsplit-1 ; k+=2)  { 
    snpnumber = (k-colbase)/2 ;

    snpnum = pedcols[snpnumber] ;
    if (snpnum<0) fatalx("logic bug (bad pedcols)\n") ;

    xvar = gvar[snpnum] ; 
    xref = gref[snpnum] ; 

    t = 0 ; 

    n = pedval(spt[k]) ; 

//  printf("zz %d %d  %d %d %d %s\n", snpnum, k, xvar, xref, n , spt[k]) ;
    if ((n<0) || (n>4)) fatalx("(getpedgenos) %s bad geno %s\n", gname, n) ;
    if (n==xvar) ++t ; 
    if ((n != xvar) && (n != xref)) t = -10 ;

    n = pedval(spt[k+1]) ; 
    if ((n<0) || (n>4)) fatalx("(getpedgenos) %s bad geno %s\n", gname, n) ;
    if (n==xvar) ++t ; 
    if ((n != xvar) && (n != xref)) t = -10 ;

    if (t<0) t = -1 ;
    cupt = snpmarkers[snpnum] ;  
    if (cupt -> ignore) continue ;
    val = t ;
    if (checkxval(cupt, indivmarkers[num], val) == NO) val = -1 ;
    putgtypes(cupt, num, val) ;
    if (val>=0) ++ngenos ;
   }
   freeup(spt, nsplit) ;
   ++num ;
  }

 free(spt) ;
 fclose(fff) ;

 for (i==0; i<nsnp; i++)  { 
   free(gcounts[i]) ;
 }

 free(gcounts) ;
 free(gref) ; 
 free(gvar) ;
 free(line) ;

 printf("genotype file processed\n") ;
 return ngenos ;
}

void
genopedcnt(char *gname, int **gcounts, int nsnp) 
{

 char *line ;
 char **spt, *sx ;
 int nsplit, num=0 ;
 int skipit ;
 int numf, snpnumber, snpnum ; 
 int k, n ;
 FILE *fff ;
 int parity, ncols, colbase ;  

// gcounts already zeroed 

 if (maxgenolinelength == -1) 
   maxgenolinelength = maxlinelength(gname) ;

//  printf("maxlinelen %d\n", maxlinelength(gname)) ;
 ZALLOC(line, maxgenolinelength+1, char) ;

 numf = 2*nsnp+10 ;
 ZALLOC(spt, numf, char *) ;

 openit(gname, &fff, "r") ;
 while (fgets(line, maxgenolinelength, fff) != NULL)  {
   nsplit = splitup(line, spt, numf) ; 
   if (nsplit == 0) continue ;
   skipit = NO ;
   sx = spt[0] ;
   if (sx[0] == '#') { 
     freeup(spt, nsplit) ;
     continue ;
   }
   if (num==0) {  
    parity = nsplit % 2 ;  
    ncols = nsplit ; 
    colbase = 6 + parity ;
   }
   for (k=colbase ; k < nsplit-1 ; k+=2)  { 
    snpnumber = (k-colbase)/2 ;
    snpnum = pedcols[snpnumber] ;
    if (snpnum<0) fatalx("logic bug (bad pedcols)\n") ;
    n = pedval(spt[k]) ; 
    if ((n<0) || (n>4)) fatalx("(genopedcnt) %s bad geno %s\n", gname, n) ;
    if (n>0) {
     ++gcounts[snpnum][n] ;
     ++num ;
    }
    n = pedval(spt[k+1]) ; 
    if ((n<0) || (n>4)) fatalx("(genopedcnt) %s bad geno %s\n", gname, n) ;
    if (n>0) {
     ++gcounts[snpnum][n] ;
     ++num ;
    }
   }
   freeup(spt, nsplit) ;
   continue ;
 }
 free(spt) ;
 free(line) ;
 fclose(fff) ;
 return  ;
}
void
outfiles(char *snpname, char *indname, char *gname, SNP **snpm, 
  Indiv **indiv, int numsnps, int numind, int packem, int ogmode) 
{
   int sizelimit = 10000000 ;
   switch (outputmode)  {  

    case EIGENSTRAT: 
     printf("eigenstrat output\n") ;
     outeigenstrat(snpname, indname, gname, snpm, indiv,  numsnps, numind)  ;
     return ;

    case PED: 
     printf("ped output\n") ;
     outped(snpname, indname, gname, snpm, indiv,  numsnps, numind, ogmode)  ;
     return ;

    case PACKEDPED: 
     printf("packedped output\n") ;
     outpackped(snpname, indname, gname, snpm, indiv,  numsnps, numind, ogmode)  ;
     return ;

   case PACKEDANCESTRYMAP:
     if (snpname != NULL) printsnps(snpname, snpm, numsnps, indiv, NO, NO) ;
     packem = YES ;
     printdata(gname, indname, snpm, indiv, numsnps, numind, packem) ;
     return ;


    default:  
     if (snpname != NULL) printsnps(snpname, snpm, numsnps, indiv, NO, NO) ;
     packem = NO ;
     if (numsnps > (sizelimit/numind)) packem = YES ;
     printdata(gname, indname, snpm, indiv, numsnps, numind, packem) ;
     return ;

   }

}

void
outeigenstrat(char *snpname, char *indname, char *gname, SNP **snpm, Indiv **indiv, int numsnps, int numind) 
{
   FILE *fff, *ifile  ;
   int  g, i, k ;
   SNP *cupt ; 
   Indiv *indx ;
   char ss[MAXSTR] ;


   settersemode(YES) ;
   if (snpname != NULL) 
    printsnps(snpname, snpm, numsnps, indiv, NO, NO) ;
  if (indname != NULL) {
   openit(indname, &ifile, "w") ;
   for(i = 0; i< numind; i++) {
      indx = indiv[i];
      if (indx->ignore) continue ;            
      strcpy(ss, indx -> egroup) ;
      if (qtmode) {  
        sprintf(ss, "%9.3f", indx -> rawqval) ;
      }
      fprintf(ifile,"%20s %c %10s",indx->ID, indx->gender,ss);
      fprintf(ifile,"\n") ;
      continue ;
   }
   fclose(ifile) ;
  }

   if (gname == NULL) return ;

   openit(gname, &fff, "w") ; 
   for (k=0; k< numsnps;  k++) {  
    cupt  = snpm[k] ;
    if (cupt -> ignore) continue ;
    if (cupt -> isfake) continue ;  
    for (i=0; i<numind; i++) {
     indx = indiv[i] ; 
     if (indx->ignore) continue ;            
     g = getgtypes(cupt, i) ;
     if (g<0) g=9 ;  
     fprintf(fff, "%1d", g);
    }
    fprintf(fff, "\n") ;
   }
   fclose(fff) ;
}
void
setgref(int **gcounts, int nsnp, int *gvar, int *gref) 
{
   int tt[5] ;  
   int i, kmax ;

   for (i=0; i<nsnp; i++) {  
    copyiarr(gcounts[i], tt, 5) ;  
    tt[0] = -9999 ;  
    ivlmaxmin(tt, 5, &kmax, NULL) ; 
    gvar[i] = kmax ;  
    if (tt[kmax] == 0) gvar[i] = 5 ;
    tt[kmax] = -9999 ; 
    ivlmaxmin(tt, 5, &kmax, NULL) ; 
    gref[i] = kmax ;  
    if (tt[kmax] == 0) gref[i] = 5 ;
   }
}
void
cleargdata(SNP **snpmarkers, int numsnps, int numindivs) 
// wipe out all genotype data 
{
   int i , k ;  
   SNP *cupt ;
   char *pbuff ;
   double y ;


   y = (double) (numindivs * 2) / (8 * (double) sizeof (char)) ;  
   rlen = nnint(ceil(y)) ;
   rlen = MAX(rlen, 48)  ;  
   packlen = rlen*numsnps ;

   if ((packmode) && (packgenos == NULL)) {
    ZALLOC(packgenos, packlen, char) ;
    clearepath(packgenos) ;
   }

   pbuff = packgenos ;


   for (i=0; i<numsnps; i++) {  
    cupt = snpmarkers[i] ;
    if (cupt -> ignore) continue ;
    if (cupt -> ngtypes == 0) { 
     if (packmode == NO) {
      ZALLOC(cupt -> gtypes, numindivs, int) ;
     }
     else  {
      ZALLOC(cupt -> gtypes, 1, int) ;
      cupt -> pbuff = pbuff ;  
      pbuff += rlen ;
     }
     cupt -> ngtypes = numindivs ;
     for (k=0; k<numindivs; ++k) {
      putgtypes(cupt, k, -1) ;
     }
    }
   }
}

void
setgenotypename(char **gname, char *iname) 
{
 if (ispedfile(iname) == NO) return ;
 if ((*gname != NULL) && strcmp(*gname, "NULL") ==0)  {  
   *gname = NULL ;  
   return ;
 }
 if (*gname != NULL) return ; 
 *gname = strdup(iname) ;
}
int maxlinelength(char *fname) 
// linelength including \n 
{

  int len, maxlen ; 
  int nl, t ; 
  FILE *fff ;

  maxlen = -1 ; 

  len = 0 ; 
  nl = (int) (unsigned char) '\n' ;

  openit(fname, &fff, "r") ;
  while ((t = fgetc(fff)) != EOF) {  
   ++len ;
   if (t==nl) {  
    maxlen = MAX(maxlen, len) ;
    len = 0 ;
   }
  }
  return maxlen ;
}
void settersemode(int mode) 
{
 tersemode = mode ;
}

void outindped(char *indname, Indiv **indiv, int numind, int ogmode)  
{  
   FILE *fff, *ifile  ;
   int  g, i, k ;
   Indiv *indx ;
   char c ;
   int pgender, astatus ;
   int dcode = 1 ;

   if (indname == NULL) return ; 

    openit(indname, &ifile, "w") ;
    for(i = 0; i< numind; i++) {
      indx = indiv[i];
      if (indx->ignore) continue ;            
      fprintf(ifile, "%6d %12s", i+1, indx->ID) ;
      fprintf(ifile, " %d %d", 0, 0) ;  // parents 
      c = indx->gender ;
      pgender = 0 ; 
      if (c == 'M') pgender = 1 ;
      if (c == 'F') pgender = 2 ;
      fprintf(ifile, " %d", pgender) ;
      if (ogmode == NO) {
       astatus = indx -> affstatus + 1 ;
       if (qtmode) {  
        fprintf(ifile, "%9.3f", indx -> rawqval) ;
       }
       else fprintf(ifile, " %d", astatus) ;
      }
      if (ogmode == YES) fprintf(ifile, " %10s", indx -> egroup) ;
      fprintf(ifile, " %d", dcode) ;
      fprintf(ifile, "\n") ;
    }
    fclose(ifile) ;
}

void
outped(char *snpname, char *indname, char *gname, SNP **snpm, Indiv **indiv, 
  int numsnps, int numind, int ogmode) 
{

   FILE *fff, *ifile  ;
   int  g, i, k ;
   SNP *cupt ; 
   Indiv *indx ;
   char c ;
   int pgender, astatus ;
   int g1, g2, dcode=1  ;

   settersemode(YES) ;
   if (snpname != NULL) 
    printmap(snpname, snpm, numsnps, indiv) ;
   if (indname!=NULL)
   outindped(indname, indiv, numind, ogmode) ;

   if (gname == NULL) return ;
    openit(gname, &fff, "w") ;
    for(i = 0; i< numind; i++) {
      indx = indiv[i];
      if (indx->ignore) continue ;            
      fprintf(fff, "%6d %12s", i+1, indx->ID) ;
      fprintf(fff, " %d %d", 0, 0) ;  // parents 
      c = indx->gender ;
      pgender = 0 ; 
      if (c == 'M') pgender = 1 ;
      if (c == 'F') pgender = 2 ;
      fprintf(fff, " %d", pgender) ;
      if (ogmode == NO) {
       astatus = indx -> affstatus +1 ;
       if (qtmode) {  
        fprintf(fff, "%9.3f", indx -> rawqval) ;
       }
       else fprintf(fff, " %d", astatus) ;
      }
      if (ogmode == YES) fprintf(fff, " %10s", indx -> egroup) ;
      fprintf(fff, " %d", dcode) ;
      for (k=0; k<numsnps; k++) { 
       cupt = snpm[k] ; 
       if (cupt -> ignore) continue ;
       if (cupt -> isfake) continue ;
       g = getgtypes(cupt, i) ;
       gtox(g, cupt -> alleles, &g1, &g2) ;
       fprintf(fff, "  %d %d", g1, g2 ) ;
      }
      fprintf(fff, "\n") ;
    }
    fclose(fff) ;
}

void gtox(int g, char *cvals, int *p1, int *p2) 
// output values for ped file using allele array
{

       int g1, g2 ;
       
       switch (g)  {
        case -1:
         *p1 = *p2 = 0; return ;
        case 0:
         g1 = 1 ; g2 = 1; break ;
        case 1:
         g1 = 1 ; g2 = 2; break ;
        case 2:
         g1 = 2 ; g2 = 2; break ;
        default:
         fatalx("(outped) bug %d\n", g) ;
       }
       if (cvals != NULL)  {  
        g1 = 3-g1 ; 
        g2 = 3-g2 ;
        g1 =  xpedval(cvals[g1-1]) ;
        g2 =  xpedval(cvals[g2-1]) ;
       }

       *p1 = MIN(g1, g2) ;
       *p2 = MAX(g1, g2) ;

}

void
outpackped(char *snpname, char *indname, char *gname, SNP **snpm, Indiv **indiv, 
  int numsnps, int numind, int ogmode) 
{

   FILE *fff, *ifile  ;
   int  g, i, k ;
   SNP *cupt ; 
   Indiv *indx ;
   char c ;
   int pgender, astatus ;
   int g1, g2, dcode=1  ;
   unsigned char ibuff[3]  ;
   unsigned char *buff ;
   int fdes, ret, blen ;
   int *gtypes ;
   double y ;


   settersemode(YES) ;
   if (snpname!=NULL)
    printmap(snpname, snpm, numsnps, indiv) ;
   if (indname != NULL)  
   outindped(indname, indiv, numind, ogmode) ;

   if (gname == NULL) return ;

   ibuff[0] = 0x6C ; 
   ibuff[1] = 0x1B ;
   ibuff[2] = 0x01 ;  

/** magic constants for snp major bed file */

   y = (double) (numind * 2) / (8 * (double) sizeof (char)) ;  
   blen = nnint(ceil(y)) ;
   ZALLOC(buff, blen, unsigned char) ;
   ZALLOC(gtypes, numind, int) ;

   fdes = open(gname, O_CREAT | O_TRUNC | O_RDWR, 0666);
 
   if (fdes<0) {
    perror("bad gname") ;
    fatalx("open failed for %s\n", gname) ;
   }
   if (verbose) 
    printf("file %s opened\n", gname) ;
 
   if (fdes<0) {
    perror("bad genoout") ;
    fatalx("open failed for %s\n", gname) ;
   }

   if (verbose) 
    printf("file %s opened\n", gname) ;

   ret = write(fdes, ibuff, 3) ;

   for (i=0; i<numsnps; i++) {  
    cupt = snpm[i] ;
    for (k=0; k<numind; ++k) { 
     gtypes[k] = getgtypes(cupt, k) ;
    }
    setbedbuff((char *) buff, gtypes, numind) ;
    ret = write(fdes, buff, blen) ;
     if (ret<0) {  
      perror("write failure") ;
      fatalx("(outpackped) bad write") ;
     } 
   }


   free(buff) ;
   close(fdes) ;
}
void
setbedbuff(char *buff, int *gtypes, int numind )
{
   int i, k ;  
   double y ;  
   int blen, wnum, wplace, bplace, t, g ;
   unsigned char c ;

   y = (double) (numind * 2) / (8 * (double) sizeof (char)) ;  
   blen = nnint(ceil(y)) ;

   c = 0xAA  ;  // missing 
   cclear((unsigned char *) buff, c, blen) ;

   for (k=0; k<numind; k++) {  
    wnum = k/4 ;
    t = k%4 ;
    wplace = 3-t ;  // switch for bed 
    bplace = 4*wnum + wplace ;
    g = bedval(gtypes[k]) ;
    wbuff((unsigned char *) buff, bplace, g) ;
   }
}
int bedval(int g) 
{
 if (g<0) return  1  ;  
 if (g==2) return 3  ;
 if (g==1) return 2 ;
 if (g==0) return 0 ;

 fatalx("(bedval) bad g value %d\n", g) ;

}

void
atopchrom(char *ss, int chrom) 
// ancestry chromosome -> map covention  
{
  switch (chrom) {  

   case 23: 
    strcpy(ss, "X") ;
    return ;

   case 24: 
    strcpy(ss, "Y") ;
    return ;

   default: 
    sprintf(ss, "%d", chrom) ;
  }
}
int
ptoachrom(char *ss) 
// map -> ancestry  
{
  char c ;

  c = ss[0] ; 

  if (c=='X') return 23 ;
  if (c=='Y') return 24 ;
  return atoi(ss) ;
}


void
printmap(char *snpname, SNP **snpm, int numsnps, Indiv **indiv) 
{ 

  char ss[5] ;
  int i ;  
  FILE *fff ;
  SNP *cupt ;
  char  c ;

  if (snpname == NULL) return  ;
  openit(snpname, &fff, "w") ;
  for (i=0; i<numsnps; i++) {  
   cupt = snpm[i] ;
   if (cupt -> ignore) continue ;
   atopchrom(ss, cupt -> chrom) ;
   fprintf(fff, "%2s", ss) ;
   fprintf(fff, " %12s", cupt -> ID) ;
   fprintf(fff, " %12.6f", cupt -> genpos) ;
   fprintf(fff, " %12.0f", cupt -> physpos) ;
   printalleles(cupt, fff) ;
   fprintf(fff, "\n") ;
  }
  fclose(fff) ;
}
char x2base(int x) 
// 12345 -> ACGTX
{
 char *blist = "?ACGT" ;
 if (x<0) return '?' ;
 if (x>4) return 'X' ;
  return blist[x] ;
}

int xpedval(char c) 
{
  char bb[2] ; 

  bb[1] = '\0' ;
  bb[0] = c ;

  if (isdigit(c)) return atoi(bb) ;
  return pedval(bb) ;



}
int pedval(char *sx) 
{
  char c ;  

  c = sx[0] ;
  if (c=='A') return 1 ;
  if (c=='C') return 2 ;
  if (c=='G') return 3 ;
  if (c=='T') return 4 ;
  if (c=='X') return 5 ;

  return (atoi(sx)) ;

}

int getbedgenos(char *gname, SNP **snpmarkers, Indiv **indivmarkers, 
 int numsnps, int numindivs, int nignore) 
{

 int val, i, k, x ;
 int t, wnum, wplace ;
 int nsnp  ;
 int ngenos = 0 ;

 SNP *cupt ;
 Indiv *indx ;

 unsigned char *buff, ibuff[3], jbuff[3] ;
 double y ;  
 int blen ;
 int fdes ;


  ibuff[0] = 0x6C ; 
  ibuff[1] = 0x1B ;
  ibuff[2] = 0x01 ;  


 cleargdata(snpmarkers, numsnps, numindivs) ;
 nsnp = numsnps + nignore ;

  y = (double) (numindivs * 2) / (8 * (double) sizeof (char)) ;  
  blen = nnint(ceil(y)) ;

  ZALLOC(buff, blen, unsigned char) ;

  fdes = open(gname, O_RDONLY) ;

  if (fdes<0) {  
   perror("open failure") ;
   fatalx("(getbedgenos) bad open %s\n", gname) ;
  } 

  t = read(fdes, jbuff, 3 ) ;

  if (t<0) {  
   perror("read failure") ;
   fatalx("(getbedgenos) bad read") ;
  } 


// check magic
 for (k=0; k<3; k++) {
  if (ibuff[k] != jbuff[k]) {
   fprintf(stderr, "magic failure: ") ;
   fprintf(stderr, " %x %x %x", jbuff[0], jbuff[1], jbuff[2]) ;
   fprintf(stderr, " %x %x %x", ibuff[0], ibuff[1], ibuff[2]) ;
   fprintf(stderr, "\n") ;
   fatalx("(getbedgenos) magic failure\n") ;
  }
 }



  for (i=0; i<nsnp; i++) {  
   cupt = snpmarkers[i] ;
   t = read(fdes, buff, blen) ; 

   if (t<0) {  
    perror("read failure") ;
    fatalx("(getbedgenos) bad read") ;
   } 

   for (k=0; k<numindivs; k++) {  
    indx = indivmarkers[k] ;
    wnum = k/4 ;
    t = k%4 ;
    wplace = 3-t ;  // switch for bed 
    wplace += 4*wnum ;
    x = rbuff(buff, wplace) ;
    val = ancval(x) ;
    if (checkxval(cupt, indx, val) == NO) val = -1 ;
    putgtypes(cupt, k, val) ;
    if (val >= 0) ++ngenos ;
   }
  }

 free(buff) ;
 printf("genotype file processed\n") ;
 return ngenos ;
}

int ancval(int x) 
// bed -> anc 
// 1/22/07 allele flipped
{
 if (x==1) return -1 ;
 if (x==3) return 0 ;
 if (x==2) return 1 ;
 if (x==0) return 2 ;
 fatalx("(ancval) bad value %d\n", x) ;
}



void setomode(enum outputmodetype *outmode, char *omode) 
{
 char *ss ;
 int len, i ;
 if (outmode == NULL) return ; 
 *outmode = ANCESTRYMAP ;
 if (omode == NULL) return ;

 ss = strdup(omode) ;
 len = strlen(ss) ;
 for (i=0; i<len ; i++) {
  ss[i] = tolower(ss[i]) ;
 }

 if (strcmp(ss, "eigenstrat") == 0)  *outmode = EIGENSTRAT ;
 if (strcmp(ss, "ascii") == 0)  *outmode = EIGENSTRAT ;
 if (strcmp(ss, "alkes") == 0)  *outmode = EIGENSTRAT ;
 if (strcmp(ss, "ped") == 0)  *outmode = PED ;
 if (strcmp(ss, "packedped") == 0)  *outmode = PACKEDPED ;
 if (strcmp(ss, "packedancestrymap") == 0)  *outmode = PACKEDANCESTRYMAP ;

 free(ss) ;
}
void snpdecimate(SNP **snpm, int nsnp, int decim, int mindis, int maxdis)  
{
 int chrom = -1 ; 
 SNP **cbuff, *cupt, *cupt2 ; 
 int k, k2, n, t  ;
  
 printf( "snpdecimate called: decim: %d  mindis: %d  maxdis: %d\n", decim, mindis, maxdis) ; 
 ZALLOC(cbuff, nsnp, SNP *) ;
 for (k=0; k<nsnp; ++k)  {  
  cupt = snpm[k] ;    
  if (cupt -> chrom != chrom)  {  
   chrom = cupt -> chrom ;
   n = 0 ;
   for (k2=k; k2 <nsnp; ++k2)   {  
    cupt2 = snpm[k2] ;
    if (cupt2 -> chrom != chrom) break ;
    if (cupt2 -> ignore) continue ;  
    if (cupt2 -> isfake) continue ;
    cbuff[n] = cupt2 ;  
    ++n ;
   }
   if (n<decim) continue ;
   decimate(cbuff, n, decim, mindis, maxdis) ;
  }
 }
}
void 
decimate(SNP **cbuff, int n, int decim, int mindis, int maxdis) 
{ 
 int k, t, u, dis, len ; 
 int *ttt ;
 SNP *cupt ; 

 cupt = cbuff[0] ;
 if (n<2) return ;
 if (n<decim) return  ;
 ZALLOC(ttt, n, int) ;
 for (k=1; k<n; ++k)  {  
  dis = (int) (cbuff[k] -> physpos - cbuff[k-1] -> physpos) ; 
  if (dis > maxdis)  {  
   decimate(cbuff, k-1, decim, mindis, maxdis) ;
   decimate(cbuff+k, n-k, decim, mindis, maxdis) ;
   return ;
  }
 }
 t = ranmod(decim) ;
 ttt[t] = 1 ;

 u = t + decim ; 

 for (;;) {  
  if (u>=n) break ;
  dis = (int) (cbuff[u] -> physpos - cbuff[t] -> physpos) ; 
   if (dis<mindis) { 
    ++u ;
    continue ; 
   }
   len = u-t-1 ;
   ivclear(ttt+t+1, 1, len) ;  
   t = u ;
   u = t + decim ;
 }
 for (k=0 ; k < n; ++k)  {  
  if (ttt[k] == 1) cbuff[k] -> ignore = YES ;
 }
// debug 
 if (verbose) {
  for (k=0 ; k < n; ++k)  {  
   printf("zz %6d %20s %20d  %3d\n", k, cbuff[k] -> ID, (int) cbuff[k] -> physpos, ttt[k]) ;
  }
 }

 free(ttt) ;

}


