#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <nicklib.h>

#include "admutils.h"
#include "qsubs.h"  

extern double a1, b1 ;
extern double thxp0, thxp1, thxp2 ;

extern Indiv **indivmarkers ; 
extern SNP **snpmarkers ;
extern int qtmode ;  
extern double *qtrho, *qtrisks ;  
extern double qtrbase ;

static int debug = NO ;

static int qnumrisks ;
static double **ccoeffs=NULL ;
static double *rcoeffs=NULL  ;
static double qmu ;


void setccr(double aa, double bb, 
  double *qrho, double *qtr, int numrisks)  
{
  int i ;  
  double qtrisk ;

  freecc() ;
  ccoeffs = initarray_2Ddouble(numrisks, 3, 0.0); 
  ZALLOC(rcoeffs, 3, double)  ; 
  qnumrisks = numrisks ;
  setqco(rcoeffs, aa, bb) ;
  qmu = aa/(aa+bb) ;

  for (i=0; i<numrisks; i++) {  
   setqcox(ccoeffs[i], aa, bb, qrho[i], qtr[i]) ;
// printf("yyy %d ",i) ;
// printmat(ccoeffs[i], 1, 3) ;
  }
}
void setqscores(Indiv *indx, double *gg, double *qsc) 
{
   double y1, y2, yran, yc, tt, ss[3], y, z, theta ;
   double *xcoeffs ;
   int i, a, num ;
   int vmode = NO ;

   num = indx -> idnum ;
   theta = indx -> theta_mode ;
   yran = qqlike(indx -> qval, rcoeffs, theta, qmu, 0) ;
   
   for (i=0; i<qnumrisks; i++) {  
    xcoeffs = ccoeffs[i] ;
    for (a=0; a<=2; a++) {  
      yc = qqlike(indx -> qval, xcoeffs, theta, qmu, a) ;
      ss[a] = yc-yran ; /* log factor if ethnicity is a */
      if (vmode) {  
       printf("zzsetq %d %d %9.3f %9.3f %9.3f\n", num, a, yc, yran, gg[a]) ;
       if (yc>20.0)  {  
         printf("zzsq %9.3f %9.3f\n", indx -> qval, theta, qmu) ;      
         printmat(rcoeffs, 1,3) ;
         printmat(xcoeffs, 1,3) ;
       }
      }
    }
    vexp(ss, ss, 3) ;
    y = vdot(ss, gg, 3) ;
    qsc[i] = log10(y) ;
  }
}

void setqco(double *qcc, double aa, double bb) 
/** 
 aa bb are Beta parameters
*/
{
 setqcox(qcc, aa, bb, qtrbase, 0.0) ;
}

void setqcox(double *qcc, double aa, double bb, 
  double qrho, double qrisk) 
/** 
 aa bb are Beta parameters
 maps qrho, qrisk to qcc.  As total variance of q is 1, we have 3 
pieces of information to fix 3 parameters.
*/
{
 double mean, var, resvar, v2, t ;  
 double c0, c1, c2 ; 

 vzero(qcc,3) ;
 if (fabs(qrho)>=1.0) fatalx("(setqcox) bad qrho %9.3f\n", qrisk) ;
 if (fabs(qrisk)>=1.0) fatalx("bad qrisk %9.3f\n", qrisk) ;

 calcbmv(aa, bb, &mean, &var) ;
 c1 = qrho/sqrt(var) ;
 resvar = (1.0-qrho*qrho) ;
 t = aa+bb ;
 v2 = 2.0*aa*bb/(t*(t+1)) ;
/* V(a-2*theta)) */
 c2 = qrisk*sqrt(resvar/v2) ;
 resvar *= (1.0-qrisk*qrisk) ;
 c0 = sqrt(resvar) ;
 qcc[0] = c0 ;
 qcc[1] = c1 ;
 qcc[2] = c2 ;
}

void calcbmv(double aa, double bb, double *mean, double *var) 
{
  double m, v, t ;

  t = aa+bb ;
  m = aa/t ;
  v = aa*bb/(t*t*(t+1)) ;
  
  *mean = m ;
  *var = v ;

}
void freecc()  
{
 if (ccoeffs != NULL) {  
  free2D(&ccoeffs, qnumrisks) ;
 }
 if (rcoeffs != NULL) free(rcoeffs) ;
}

double genqval(double *qcc, double theta, double mu, int a) 
{
   double y, z ; 
  
   y = qcc[0]*gauss() ;
   y += qcc[1]*(theta-mu) ;
   z = (double) a -2.0*theta ;
   y += qcc[2]*z ;
   return y ;
}
double lnordens(double xval, double sig) 
/* log likelihood for xval mean 0 s.dev. sig */  
{
  double y, t ;

  y = -log(sig) ;
  t = xval/sig ;
  y -= 0.5*t*t ;
  return y ;
}

double qqlike(double qval, double *cc, double theta, double mu, int a)
/* log likelihood of qval */
{
  double y, z ;
  y = qval ; 
  y -= cc[1]*(theta-mu) ; 

  if (a<0) fatalx("bad qqlike %d\n",a) ;
  z = ((double) a) - 2.0*theta ;
  y -= cc[2]*z ;

  return lnordens(y, cc[0]) ;

}
void doprobit(Indiv **indm, int numindivs) 
{
  double *aq, *bq ;  
  int i, n ;
  Indiv *indx ;

  ZALLOC(aq, numindivs, double) ;
  ZALLOC(bq, numindivs, double) ;
  n = 0 ;
  for (i=0; i<numindivs; i++) {  
   indx = indm[i] ;
   if (indx -> ignore) continue ;
   aq[n] = indx -> rawqval ;
   ++n ;
  }
  probit(bq, aq, n) ;
  n = 0 ;
  for (i=0; i<numindivs; i++) {  
   indx = indivmarkers[i] ;
   if (indx -> ignore) continue ;
   indx -> qval  = bq[n] ;
   ++n ;
  }



 free(aq) ;
 free(bq) ;

}

void probit(double *xout, double *xin, int n) 
/** probit transform */
{
   int i ;
   double tail, y, top, bot ; 
   double *a ;
   int *ind ;

   ZALLOC(a, n, double) ;
   ZALLOC(ind, n, int) ;

   vst(a, xin, -1.0, n) ;
   sortit(a, ind, n) ;
   bot = (double) (n+1) ;
   for (i=0; i<n; i++)  {
    top = (double) (i+1) ;
    tail = top/bot ;
    y = zprob(tail) ;
    xout[ind[i]] = y ;
   }

/**
   for (i=0; i<n; i++)  {
    printf("zz %4d %9.3f %9.3f\n", i, xin[i], xout[i]) ;
   }
*/

   free(a) ;
   free(ind) ;


}

double qqltot(Indiv **indm, int numindivs, double *cco, double mu) 
{
     Indiv *indx ;
     double theta, y, qval ;
     int i ;

     y = 0.0 ;
     for (i=0; i<numindivs; i++)  {  
      indx = indivmarkers[i] ;
      if (indx->ignore) continue ;
      theta = indx -> theta_mode ;
      qval = indx -> qval ;
      y += qqlike(qval, cco, theta, mu, 0) ;
     }
  return y ;
}

