#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <nicklib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "getpars.h"

#define  VERSION  "1.00"

double C= 23, G=33, lambdal = 4, lambdaa=6, r1 = 2, r2 = -1 ;
double theta = 0.2 ;
double p=0.2, pp=0.8 ;
double nassoc = 1.0e6  ;  
double fpos = .05 ;
double fneg = .20 ;
double dfreq = 0.0 ; /* disease prob */

int usepp = NO ;
int useadmix = YES ;
int twosided = NO ;
int verbose = NO ;

double ss[3] ;
double eee ;

double assocsamp, linksamp, ancestsamp ;

void readcommands(int argc, char **argv) ;

double 
dolinkage(double  *v, 
 double lambda, double fpos, double fneg)   ;

double 
doancest(double theta, 
 double  psi2,  
 double lambda, double fpos, double fneg)  ;

double klsig(double C, double G, double lambda, double T)  ;
double klsiginv(double C, double G, double lambda, double mu) ;
double knewt(double C, double G, double lambda, double z, double ptail);

void bivar_srate(double r, double q, double *ym, double *yv) ;
double  klbinv(double pr, double C, double G, double lambda, 
 double tc, double sig, double lexp)  ;
double  klb(double nsamp, double C, double G, double lambda, 
 double tc, double sig, double lexp)  ;
double  klbd(double nsamp, double C, double G, double lambda, 
 double tc, double sig, double lexp)  ;
double klsigd1(double C, double G, double lambda, double T) ;

void sratex(double m, double v, double *ym, double *yv) ;

double bayeskl(double muc, double vc, double C, double G, 
 double lambda, double fpos, double fneg) ;


int main(int argc, char **argv)
{
 double psi, psi2 ;

 assocsamp = linksamp = ancestsamp ;
 readcommands( argc, argv)  ;

 linksamp =  dolinkage(ss, lambdal, fpos, fneg) ;
 if (verbose)  {
  printf("linksamp: %9.3f\n", linksamp) ;
  printf("\n\n") ;
 }

 psi2 = ss[0]*eee/asum(ss,3) ;
 if (verbose)
  printf("psi2: %9.3f\n", psi2) ;
 ancestsamp = doancest(theta, psi2, lambdaa, fpos, fneg) ;
 if (verbose) 
  printf("ancestsamp: %9.3f\n", ancestsamp) ;
 if (!verbose) printf("%9.3f %9.3f \n", 
   linksamp, ancestsamp) ;
}

double 
dolinkage(double  *vv, 
 double lambda, double fpos, double fneg)  

{
   double tt[3], ttr[3], y, qq ;
   double tibd[4], tcibd[4], tmat[9] ;
   double pdd[3], risk[3], pr[3], pc[3] ;
   double var[3] ;  
   double meanr[3], mvar[3], sss[3], prior[3] ; 
   double mur, vr, muc, vc, fom, g, csig, rsig;
   double y1, y2, rbase, ns ;
  
   int l1, l2, x, i ;

   y = asum(vv, 3) ; 
   vst(var, vv, 1.0/y, 3) ;
   if (verbose) {
    printf("linkage: \n\n") ;
    for (i=0; i<3; i++) {  
     printf("variance component: %2d %9.3f\n",i, var[i]) ;
    }
   }
   rbase = 0.5*(var[0]+var[1]) ;  /* null correlation for sibs */
   sss[0] = 0.5*var[1] ;
   sss[1] = sss[0] + 0.5*var[0] ;
   sss[2] = sss[1] + 0.5*var[0] ;
   prior[0] = prior[2] = 0.25 ;  prior[1] = 0.5 ;
/* 
 these are the correlations for IBD 0, 1, 2 
*/
  for (i=0; i<3; i++)  {  
   bivar_srate(sss[i], rbase, meanr+i, mvar+i) ;
  }
  if (verbose) {
   printf("\n\n") ;
   printf("%3s %12s %12s %12s\n",     
    "IBD", "rho", "mean rate",  "variance") ;
   for (i=0; i<3; i++)  {  
    printf("%3d %12.6f %12.6f %12.6f\n",     
     i, sss[i], meanr[i], mvar[i]) ;
   }
   printf("\n\n") ;
  }
  y1 = vdot(prior, meanr, 3) ;

  vvt (tt, meanr, meanr, 3) ;
  vvp( tt, tt, mvar, 3) ;

  y2 = vdot(prior, tt, 3) ;
  muc = y1 ;
  vc =  y2 - y1*y1 ;  

  if (verbose) {
   printf("scoring rate: %12.6f\n", muc)  ;
   printf("    variance: %12.6f\n", vc)  ;
  }

   ns = bayeskl(muc, vc, C, G, lambda, fpos, fneg) ;
   return ns ;

}
double bayeskl(double muc, double vc, double C, double G, 
 double lambda, double fpos, double fneg) 
{
  double lexp, cz, ns ;

   lexp = 0.5*(vc-2*muc) ;
   if (verbose) 
    printf("lexp: %12.6f\n", lexp) ;

   cz = zprob(1-fneg) ; 
   ns = klbinv(fpos, C, G, lambda, cz, sqrt(vc), lexp) ;


  return ns ;


}

double 
doancest(double theta, 
 double  psi2,  
 double lambda, double fpos, double fneg)  

/** 
 psi2 is variance due to ethnicity at locus
*/

{
   double tt[3], ttr[3], y, qq, y1, y2, c, c2 ;
   double ta[2], tb[2] ;
   double tmat[3] ;
   double prior[3], ee[3], e2[3], meanr[3], mvar[3] ;
   double muc, vc, fom, g, sig, lexp, cz, ns;
   int l1, l2, x, i ;

   sett1(prior, theta, 3) ;
   for (i=0; i<3; i++) 
    ee[i] = i ;
   y = vdot(prior, ee, 3) ; 
   vsp(ee, ee, -y, 3) ;  /* take off mean */
   vvt(e2, ee, ee, 3) ;  
   y = vdot(prior, ee, 3) ; 
   y2 = vdot(prior, e2, 3) ; 
   c2 = psi2/y2 ;
   c = sqrt(c2) ;  
// c * ee   gives contribution to psi2
   qq = 1.0-psi2 ;

  vst(ee, ee, c, 3) ;
  for (i=0; i<3; i++)  {  
   sratex(ee[i],qq, meanr+i, mvar+i) ;
  }
  if (verbose) {
   printf("ancestry: \n\n") ;
   printf("theta: %9.3f\n\n", theta) ;
   printf("%3s %9s %9s   %12s %12s\n",     
    "E",  "EE", "res. var.", "mean rate",  "variance") ;
   for (i=0; i<3; i++)  {  
    printf("%3d %9.3f %9.3f   %12.6f %12.6f\n",     
     i, ee[i], qq, meanr[i], mvar[i]) ;
   }
   printf("\n\n") ;
  }
   
  vvt (tt, meanr, meanr, 3) ;
  vvp( tt, tt, mvar, 3) ;

  y2 = vdot(prior, tt, 3) ;
  muc = y1  = vdot(prior, meanr, 3) ;
  vc =  y2 - y1*y1 ;  

  if (verbose) {
   printf("scoring rate: %12.6f\n", muc)  ;
   printf("    variance: %12.6f\n", vc)  ;
  }

  ns = bayeskl(muc, vc, C, G, lambda, fpos, fneg) ;
  return ns ;

}

void readcommands(int argc, char **argv) 

{
  char *parname = NULL ;
  phandle *ph ;
  char c;

  while ((c = getopt (argc, argv, "p:v")) != -1) {

    switch (c)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
        printf("version: %s\n", VERSION) ;
	parname = strdup(optarg) ;
	break;

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   ph = openpars(parname) ;

   getdbl(ph,"C:", &C) ;
   getdbl(ph,"G:", &C) ;
   getdbl(ph,"lambdal:", &lambdal) ;
   getdbl(ph,"lambdaa:", &lambdaa) ;
   getdbl(ph,"theta:", &theta) ;
   getdbl(ph,"p:", &p) ;
   getdbl(ph,"pp:", &pp) ;
   getdbl(ph,"fpos:", &fpos) ;
   getdbl(ph,"fneg:", &fneg) ;
   getdbls(ph, "linkagevar:", ss,  3) ;
   getdbl(ph, "ee0:", &eee) ;
   useadmix = YES ;
   getint(ph,"verbose:", &verbose) ;
}

double klsiginv(double C, double G, double lambda, double mu) 
/** 
 return thresghold T so that random KL prob is mu
*/
{
  double t, tx ;
  if (mu>0.5) 
   return 0.0 ;
  if (mu <= 0.0) return 50 ;
  tx = 0.0 ;  
  for (;;) {  
   ++tx ;
   if (klsig(C,G, lambda, tx) < mu) { 
    tx -= 0.5 ; break ;
   }
  }
  t = knewt(C, G, lambda, tx, mu) ;
  return t ;
}
double knewt(double C, double G, double lambda, double z, double ptail)
{
/**
 newton step
 z is approximation
*/
    double p0, pder, h ;
    double pi, zz ;
    double a, b, ader, bder, xa, xb, ltail ;
    int iter ;

    pi = 2.0*acos(0.0) ;
    zz = z ;
    ltail = log(ptail) ;
    
    for (iter = 1; iter<=10; ++iter) {
     p0 = klsig(C, G, lambda, zz) ; p0 = log(p0) ;
/**
     printf("%g %g %g %20.10f\n", zz, p0, ltail, ltail-p0) ;
*/
     xa = ntail(zz) ; a = log(xa) ;
     xb = C + lambda*G*zz*zz ; b = log(xb) ;
     ader = -exp(-0.5*zz*zz)/(sqrt(2*pi)*xa) ;
     bder = 2*lambda*G*zz/xb ;
     pder = bder + ader ;
     if (pder==0.0) return zz ;
     h = (ltail-p0)/pder ;
     if (h==0.0) return zz ;
     zz += h ;
    }
    return zz ;
}
void
 mkx(double *x, double *a, double *b)        
{

  x[0] = a[0]*b[0] ;
  x[1] = a[1]*b[0] + a[0]*b[1] ;
  x[2] = a[1]*b[1] ;

}

void bivar_srate(double r, double q, double *ym, double *yv) 
/** 
 scoring rate for bivariate normal 
 each component standard 
 correlation r vs q 
 ym is mean yvar variance per observation (r is true) 
*/
{
  double ysc, yvar ;
  double aa[4], bb[4], y ;
  double ll[4], ltran[4], a, b ;
  double w1[4], w2[4], c0, c1, d  ;
  *ym = *yv = 0.0 ;
  if (r==q) return ;
  ysc = -0.5 * log(1-r*r) ;
  ysc -= -0.5 * log(1-q*q) ;
 
  y = -0.5/(1-r*r) ;
  aa[0] = aa[3] = 1.0 ;
  aa[1] = aa[2] = -r ;
  vst(aa, aa, y, 4) ;  /* quadratic form */

  y = -0.5/(1-q*q) ;
  bb[0] = bb[3] = 1.0 ;
  bb[1] = bb[2] = -q ;
  vst(bb, bb, y, 4) ;  /* quadratic form */
  vvm(aa, aa, bb, 4) ;  
  a = r ; b = sqrt(1-r*r); 
  /* x2 = a*x1 + b * n    n standard normal */
  ll[0] = 1 ; ll[1] = 0; ll[2] = a ; ll[3] = b ;
  transpose(ltran, ll, 2, 2) ;
  mulmat(w1, ltran, aa, 2, 2, 2) ;
  mulmat(w2, w1, ll,  2, 2, 2) ;
//printmat(w2, 2, 2) ;  /* q form on independent normals */
  c0 = w2[0] ; c1 = w2[3], d = w2[1]  ;
  ysc += (c0+c1) ;
  yvar = 2*c0*c0+2*c1*c1+4*d*d ; 
  *ym = ysc ;
  *yv = yvar ;
}

/** 

 the following little package 
 do a Kruglyak-Lander analysis when the statistic 
 of interest is a log bayes factor.  
 tc is the threshold "sigmage" 
 for the causal distribution  

 So 
  tc = zprob(1-fneg) ;
  pr is false positive rate.  
  sig is s.dev / unit observation.  
  lexp is a correction.   
  If y is a score N( N mu, N var) in causal case 
  then under the null 
  P_R(y) = exp(-y) * P_C(y) 
  which leads in the region of interest to the 
  KL theory 
  pr = exp(-l N) KL(tr)   
   where 
   tr = tc + sqrt(nsamp) * sig ;
  we solve this for N in klbinv
  klb gives the KL tail area 
*/

double  klbinv(double pr, double C, double G, double lambda, 
 double tc, double sig, double lexp)  
{
  int iter, niter=0 ;
  double ns, y, yd, h, z, zd ;
  ns = 2.0 ;
  for (;;) {  
   ++niter ;
   if (niter>1000) fatalx("(klbinv) too many iters\n") ;
   y = klb(ns, C, G, lambda, tc, sig, lexp) ;
   if (y<pr) break ;
   ns *= 2.0 ;
  }
  ns /= 2.0 ;
  for (iter=1; iter <= 10; ++iter) { 
   y = klb(ns, C, G, lambda, tc, sig, lexp) ;
   yd = klbd(ns, C, G, lambda, tc, sig, lexp) ;
   z = log(y) ;
   zd = yd/y ;
   h = (log(pr) - z)/ zd ; 
// printf("%3d %15.9f %15.9f %15.9f %15.9f\n",iter, log(pr), z, ns, h) ;
   ns += h ;
  }

  if (verbose) {  
   printf("threshold (causal): %9.3f\n", tc) ;
   printf("threshold (random): %9.3f\n", tc + sqrt(ns)*sig) ;
  }

  return ns ;
}

double  klb(double nsamp, double C, double G, double lambda, 
 double tc, double sig, double lexp)  
{
 double tr, klrawsig ;
 tr = tc + sqrt(nsamp) * sig ;
 klrawsig = klsig(C, G, lambda, tr) ;
 return exp(lexp*nsamp)*klrawsig ;
}

double  klbd(double nsamp, double C, double G, double lambda, 
 double tc, double sig, double lexp)  
{
 double tr, klrawsig, d1, ld1, y ;
 tr = tc + sqrt(nsamp) * sig ;
 klrawsig = klsig(C, G, lambda, tr) ;
 y = exp(lexp*nsamp)  ;
 ld1 = lexp*y ;
 d1 = klsigd1(C,G,lambda,tr)*(0.5*sig)/sqrt(nsamp) ;
 return ld1*klrawsig + y*d1 ;
}


double klsig(double C, double G, double lambda, double T) 
{
   double alpha, mu  ;
   alpha = ntail(T) ;
   mu = (C + lambda*G*T*T)*alpha ;
   return mu ;
}
double klsigd1(double C, double G, double lambda, double T) 
{
   double alpha, mu  ;
   double ad1, xd, deriv ; 

   alpha = ntail(T) ;
   ad1 = -nordis(T) ;
   xd = 2*lambda*G*T ;
   
   deriv = (C + lambda*G*T*T)*ad1 + xd*alpha ;
   return deriv ;
}
void sratex(double m, double v, double *ym, double *yv) 
/** 
 scoring rate: causal N(m,v) 
               random N(0,1) 
*/
{
  double cc[3], dd[3], ysc, a, b, var  ;
/** 
 x is observation
 set z = (x-m)/s  So that z is standard normal (causal) 
 then at random z = au + b where a=1/s b = m/s  
*/
 a = 1/sqrt(v) ;
 b = m/sqrt(v) ;
 ysc = log(a) ;
 cc[0] = cc[1] = 0.0 ;
 cc[2] = 1.0 ;
 vst(cc, cc, -0.5, 3) ;
 dd[0] = b*b ;
 dd[1] = -2*b ;  
 dd[2] = 1.0 ;
 vst(dd, dd, -0.5/(a*a), 3) ;
 vvm(cc, cc, dd, 3) ;  
/* 
 score is  
 ysc + cc[0] + cc[1]*z + cc[2] *z*z 
 where z is standard normal
*/
 ysc += cc[0] + cc[2] ;  /* mean */
 var = cc[1]*cc[1] + 2.0*cc[2]*cc[2] ;  
 *ym = ysc ; 
 *yv = var ;
}

